/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*-
 *
 * Copyright (C) 2008 Richard Hughes <richard@hughsie.com>
 * Copyright (C) 2009 Leandro Pereira <leandro@hardinfo.org>
 * Copyright (C) 2015 James B
 * Copyright (C) 2016, 2023 step
 *
 * Licensed under the GNU General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
		/* disable G_ADD_PRIVATE macro deprecation warning introduced in 2.58 */
#define GLIB_VERSION_MIN_REQUIRED	GLIB_VERSION_2_56
#include <glib.h>
#include <libintl.h>
#include <ctype.h>

//#define DEBUG g_print
#define DEBUG

#include "egg-markdown.h"

/*******************************************************************************
 *
 * This is a simple Markdown parser.
 * It can output to Pango, HTML, ANSI codes, or plain text.
 *
 * It extends over basic markdown support to:
 * - auto-replace certain word patterns, like http://..., with code spans
 * - auto-replace smart quote and m-dash
 * - accept embedded HTML comments
 * - implement fenced code blocks
 * - display markdown pipe tables as code blocks
 * - extract or replace gettext translation data
 * - execute embedded code.
 *
 * The following limitations are already known, and properly deliberate:
 *
 * - No ordered list support
 * - No blockquote section support
 * - No backslash escapes support
 * - No HTML escaping support
 *
 * It does support the rest of the standard pretty well, although it's not
 * been run against any conformance tests. The parsing is single pass, with
 * a simple enumerated interpreter mode and a single line back-memory.
 *
 ******************************************************************************/

/*******************************************************************************
 * We implement some text formatting operations using Unicode Private Use Area
 * (PUA) codepoints. Input markdown shall not include the PUA characters we use.
 ******************************************************************************/

/**
 * https://www.unicode.org/faq/private_use.html
 * https://github.com/silnrsi/unicode-resources/tree/main/sil-pua
 **/

/*
* A markdown hard line-break (2+ spaces at the end of a line) is stored in
* back-memory as iUNIPUA_BR, and presented to gettext as MSGSTR_BR.
* See egg_markdown_to_text_line_formatter_exec.
*/
#define iUNIPUA_BR         0xF48A         /* EF928A */
#define sUNIPUA_BR         "\357\222\212"
#define MSGSTR_BR               "<%BR/%>"
/*
* Markdown code spans are stored internally as sUNIPUA_CODE.
*/
#define iUNIPUA_CODE       0xF489
#define sUNIPUA_CODE       "\357\222\211"
/*
* Markdown link and image spans are stored internally as sUNIPUA_LINK.
*/
#define iUNIPUA_LINK       0xF488
#define sUNIPUA_LINK       "\357\222\210"

/*
* Input lines are joined together in back-memory self->priv->pending separated
* by PENDING_SEPARATOR (egg_markdown_add_pending) - chosen to be a space) " ".
* Space separated text is easy to parse. If the separator included non-space
* section- or word-boundary characters, parsing would be more complex.
*/
#define PENDING_SEPARATOR " "
/* make it stand out while debugging */
//#define PENDING_SEPARATOR "┼"

/*******************************************************************************
 ******************************************************************************/

static void     egg_markdown_finalize	(GObject		*object);

#define EGG_MARKDOWN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), EGG_TYPE_MARKDOWN, EggMarkdownPrivate))

typedef gchar *(EggMarkdownLinkBuilder)(const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id);
typedef gchar *(EggMarkdownImageBuilder)(const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id);

typedef enum {
	EGG_MARKDOWN_MODE_BLANK,
	EGG_MARKDOWN_MODE_RULE,
	EGG_MARKDOWN_MODE_BULLETT,
	EGG_MARKDOWN_MODE_PARA,
	EGG_MARKDOWN_MODE_CODEBLOCK,
	EGG_MARKDOWN_MODE_H1,
	EGG_MARKDOWN_MODE_H2,
	EGG_MARKDOWN_MODE_H3,
	EGG_MARKDOWN_MODE_COMMENTBLOCK,
	EGG_MARKDOWN_MODE_UNKNOWN
} EggMarkdownMode;

typedef enum {
	EGG_MARKDOWN_CODEBLOCK_NONE,
	EGG_MARKDOWN_CODEBLOCK_STD,
	EGG_MARKDOWN_CODEBLOCK_SHEBANG,
	EGG_MARKDOWN_CODEBLOCK_FENCED,
	EGG_MARKDOWN_CODEBLOCK_TABLE,
} EggMarkdownCodeblockMode;

typedef enum {
	EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE,
	EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT
} EggMarkdownCodeblockSyntax;

typedef struct {
	const gchar *em_start;
	const gchar *em_end;
	const gchar *strong_start;
	const gchar *strong_end;
	const gchar *strong_em_start;
	const gchar *strong_em_end;
	const gchar *code_start;
	const gchar *code_end;
	const gchar *codeblock_start;
	const gchar *codeblock_end;
	const gchar *codeblock_table_start;
	const gchar *codeblock_table_end;
	const gchar *codeblock_shebang_start;
	const gchar *codeblock_shebang_end;
	const gchar *strikethrough_start;
	const gchar *strikethrough_end;
	const gchar *h1_start;
	const gchar *h1_end;
	const gchar *h2_start;
	const gchar *h2_end;
	const gchar *h3_start;
	const gchar *h3_end;
	const gchar *bullett_start;
	const gchar *bullett_end;
	const gchar *rule;
	const gchar *para_start;
	const gchar *para_end;
	const gchar *br;

	EggMarkdownLinkBuilder *link_builder;
	EggMarkdownImageBuilder *image_builder;
} EggMarkdownTags;

struct EggMarkdownPrivate
{
	EggMarkdownExtensions      extensions;
	EggMarkdownMode            mode;
	EggMarkdownMode            prev_mode;
	EggMarkdownCodeblockMode   codeblock_mode;
	EggMarkdownCodeblockSyntax codeblock_syntax;
	gint                       codeblock_start_line_idx;
	gint                       curr_line_idx;     // input lines
	gboolean                   eof;               // no more input lines
	EggMarkdownTags            tags;
	EggMarkdownOutput          output;
	gint                       max_lines;
	guint                      line_count;        // count of output lines
	gboolean                   smart_quoting;
	gboolean                   escape;
	gboolean                   autocode;
	gboolean                   autocode_shebang;  // autocode entire file that has shebang
	gboolean                   exec;              // run executable code
	gboolean                   global_exec;
	gboolean                   use_gettext;       // translate content via gettext
	GString                    *pending;
	GString                    *processed;
	GArray                     *regex_table;
	GArray                     *code_table;
	GArray                     *link_table;

	// for pot_output
	gchar*                     input_filename;
	FILE*                      pot_output;        // where to send pot output
	gboolean                   temporary_disable_pot;

	// code blocks mode save
	gboolean                   codeblock_prev_exec;
	gboolean                   codeblock_prev_pot;

	// code block table mode
	guint                      table_columns;
};

/**
 * EggMarkdownRegexType
 **/
typedef enum {
	EGG_MARKDOWN_REGEX_CODE = 0,        /* md code spans */
	EGG_MARKDOWN_REGEX_CODE_REF,        /* internal code_refs */
	EGG_MARKDOWN_REGEX_LINK,            /* md link and image element */
	EGG_MARKDOWN_REGEX_LINK_DEST_REF,   /* internal link_dest_ref for internet URI and image path */
	EGG_MARKDOWN_REGEX_LINK_TITLE_REF,  /* internal_title_ref for link and image title */
	EGG_MARKDOWN_REGEX_LEN,             /* length of regex_table */
} EggMarkdownRegexType;

/**
 * EggMarkdownSliceType
 **/
typedef enum {
	/* any match */
	EGG_MARKDOWN_SLICE_MATCH = 0,     /* text of a whole match */
	EGG_MARKDOWN_SLICE_BEFORE,        /* text before a MATCH */
	EGG_MARKDOWN_SLICE_AFTER,         /* text after the last MATCH */
	EGG_MARKDOWN_SLICE_IGNORED,
	EGG_MARKDOWN_SLICE_UNKNOWN,

	/* code spans */
	EGG_MARKDOWN_SLICE_CODE_CODE,      /* code between backticks */
	EGG_MARKDOWN_SLICE_CODE_MID,       /* text before optional TICK and CODE */
	EGG_MARKDOWN_SLICE_CODE_TICK,      /* if exactly 1 space, it signals that CODE includes backtick(s) */
	EGG_MARKDOWN_SLICE_CODE_END,       /* text after CODE and optional TICK */

	/* links */
	EGG_MARKDOWN_SLICE_LINK_TYPE_URL,  /* "[" */
	EGG_MARKDOWN_SLICE_LINK_TYPE_IMG,  /* "![" */
	EGG_MARKDOWN_SLICE_LINK_TEXT,      /* text between square brackets */
	EGG_MARKDOWN_SLICE_LINK_DEST,      /* URL or path between brackets */
	EGG_MARKDOWN_SLICE_LINK_TITLE,     /* optional accessibility title */
	EGG_MARKDOWN_SLICE_LINK_MID,       /* text before the next TYPE or between TEXT and DEST or between DEST and TITLE */
	EGG_MARKDOWN_SLICE_LINK_END,       /* text after the CODE and optional TITLE */
} EggMarkdownSliceType;

/* debugging helpers C99 */
//#define PRINTERS	0
//#define PRINTERS	1
//#define PRINTERS	2
#undef  PRINTERS
#if PRINTERS > 0
static const gchar * const _slice_type_name[] = {
	[EGG_MARKDOWN_SLICE_MATCH]         = "MATCH",
	[EGG_MARKDOWN_SLICE_BEFORE]        = "BEFORE",
	[EGG_MARKDOWN_SLICE_AFTER]         = "AFTER",
	[EGG_MARKDOWN_SLICE_IGNORED]       = "IGNORED",
	[EGG_MARKDOWN_SLICE_UNKNOWN]       = "UNKNOWN",

	[EGG_MARKDOWN_SLICE_CODE_CODE]     = "CODE_CODE",
	[EGG_MARKDOWN_SLICE_CODE_MID]      = "CODE_MID",
	[EGG_MARKDOWN_SLICE_CODE_TICK]     = "CODE_TICK",
	[EGG_MARKDOWN_SLICE_CODE_END]      = "CODE_END",

	[EGG_MARKDOWN_SLICE_LINK_TYPE_URL] = "LINK_TYPE_URL",
	[EGG_MARKDOWN_SLICE_LINK_TYPE_IMG] = "LINK_TYPE_IMG",
	[EGG_MARKDOWN_SLICE_LINK_TEXT]     = "LINK_TEXT",
	[EGG_MARKDOWN_SLICE_LINK_DEST]     = "LINK_DEST",
	[EGG_MARKDOWN_SLICE_LINK_TITLE]    = "LINK_TITLE",
	[EGG_MARKDOWN_SLICE_LINK_MID]      = "LINK_MID",
	[EGG_MARKDOWN_SLICE_LINK_END]      = "LINK_END",
};
#endif /* PRINTERS */

static const gchar * const _tag_info[] = {
	[EGG_MARKDOWN_TAG_DEST_LINK_URI_ID]   = "dest=Lu",
	[EGG_MARKDOWN_TAG_DEST_IMAGE_PATH_ID] = "dest=Ip",
};

/**
 * egg_markdown_tag_get_info:
 * @tag: pango "font" tag
 * @subject: EggMarkdownTagInfo
 * Return: `gint` @subject's value parsed from @tag
 **/
gint
egg_markdown_tag_get_info (EggMarkdown *self, const gchar *tag, const EggMarkdownTagInfo subject) {
	gchar *p = NULL;
	gint ret  = -1;

	g_return_val_if_fail (self != NULL, -1);
	g_return_val_if_fail (tag  != NULL, -1);
	g_return_val_if_fail (subject < EGG_MARKDOWN_TAG_INFO_LEN, -1);
	p = strstr (tag, _tag_info[subject]);
	if (p)
		ret = atoi (p + strlen (_tag_info[subject]));
	return ret;
}

/**
 * EggMarkdownSlice
 * String slice specialized for markdown text
 **/
typedef struct {
	EggMarkdownSliceType type;
	gint    start;
	gint    len;   /* bytes */
	gint    data;
} EggMarkdownSlice;

G_DEFINE_TYPE (EggMarkdown, egg_markdown, G_TYPE_OBJECT)

static inline void
skip_blanks (gchar **p) {
	while (**p == ' ' || **p == '\t') (*p)++;
}

/**
 * egg_markdown_to_text_line_is_rule:
 *
 * Horizontal rules are created by placing three or more hyphens, asterisks,
 * or underscores on a line by themselves.
 * You may use spaces between the hyphens or asterisks.
 **/
static gboolean
egg_markdown_to_text_line_is_rule (const gchar *line)
{
	guint i;
	guint len;
	guint count = 0;
	gchar *copy = NULL;
	gboolean ret = FALSE;

	len = strnlen (line, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* replace non-rule chars with ~ */
	copy = g_strdup (line);
	g_strcanon (copy, "-*_ ", '~');
	for (i=0; i<len; i++) {
		if (copy[i] == '~')
			goto out;
		if (copy[i] != ' ')
			count++;
	}

	/* if we matched, return true */
	if (count >= 3)
		ret = TRUE;
out:
	g_free (copy);
	return ret;
}

/**
 * egg_markdown_to_text_line_is_bullett:
 **/
static gboolean
egg_markdown_to_text_line_is_bullett (const gchar *line)
{
	return (g_str_has_prefix (line, "- ") ||
		g_str_has_prefix (line, "* ") ||
		g_str_has_prefix (line, "+ ") ||
		g_str_has_prefix (line, " - ") ||
		g_str_has_prefix (line, " * ") ||
		g_str_has_prefix (line, " + "));
}

/**
 * egg_markdown_to_text_line_is_header1:
 **/
static gboolean
egg_markdown_to_text_line_is_header1 (const gchar *line)
{
	return g_str_has_prefix (line, "# ");
}

/**
 * egg_markdown_to_text_line_is_header2:
 **/
static gboolean
egg_markdown_to_text_line_is_header2 (const gchar *line)
{
	return g_str_has_prefix (line, "## ");
}

/**
 * egg_markdown_to_text_line_is_header3:
 **/
static gboolean
egg_markdown_to_text_line_is_header3 (const gchar *line)
{
	return g_str_has_prefix (line, "### ");
}


/**
 * egg_markdown_to_text_line_is_header1_type2:
 **/
static gboolean
egg_markdown_to_text_line_is_header1_type2 (const gchar *line)
{
	return g_str_has_prefix (line, "===");
}

/**
 * egg_markdown_to_text_line_is_header2_type2:
 **/
static gboolean
egg_markdown_to_text_line_is_header2_type2 (const gchar *line)
{
	return g_str_has_prefix (line, "---");
}

/**
 * egg_markdown_to_text_line_is_textdomain:
 **/
static gboolean
egg_markdown_to_text_line_is_textdomain (const gchar *line)
{
	return g_str_has_prefix (line, TEXTDOMAIN_DIRECTIVE);
}

/**
 * egg_markdown_to_text_line_is_nopot:
 **/
static gboolean
egg_markdown_to_text_line_is_nopot (const gchar *line)
{
	return g_str_has_prefix (line, NOPOT_DIRECTIVE);
}

/**
 * egg_markdown_to_text_line_is_exec:
 **/
static gboolean
egg_markdown_to_text_line_is_exec (const gchar *line)
{
	return g_str_has_prefix (line, EXEC_DIRECTIVE);
}

/**
 * egg_markdown_to_text_line_is_shebang:
 **/
static gboolean
egg_markdown_to_text_line_is_shebang (const gchar *line)
{
	return g_str_has_prefix (line, "#!");
}

/**
 * egg_markdown_to_text_line_is_indented_code
 **/
static gboolean
egg_markdown_to_text_line_is_indented_code (const EggMarkdown *self, const gchar *line)
{
	/* https://spec.commonmark.org/0.30/#indented-code-blocks */

	/* spec: "An indented code block cannot interrupt a paragraph, so there must
	be a blank line between a paragraph and a following indented code block." */

	/* spec: "Any non-blank line with fewer than four spaces of indentation ends
	the code block. A paragraph may occur immediately after indented code." */

	gboolean ret = self->priv->mode != EGG_MARKDOWN_MODE_PARA;
	if (ret)
		ret = (g_str_has_prefix (line, "    ") || g_str_has_prefix (line, "\t"));
	return ret;
}

static gboolean
egg_markdown_to_text_line_is_fenced_code (const gchar *line, EggMarkdownCodeblockSyntax *syntax)
{
	if( g_str_has_prefix (line, "```") ) {
		if( strncmp(line + 3, "gettext", 7)==0 )
			*syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT;
		else
			*syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
		return TRUE;
	}
	return FALSE;
}

static gboolean
egg_markdown_extension_enabled(EggMarkdown *self, const EggMarkdownExtensions ext) {
	return (self->priv->extensions & ext);
}

#if 0
/**
 * egg_markdown_to_text_line_is_blockquote:
 **/
static gboolean
egg_markdown_to_text_line_is_blockquote (const gchar *line)
{
	return (g_str_has_prefix (line, "> "));
}
#endif

/**
 * egg_markdown_to_text_line_is_blank:
 **/
static gboolean
egg_markdown_to_text_line_is_blank (const gchar *line)
{
	guint i;
	guint len;
	gboolean ret = TRUE;

	len = strnlen (line, EGG_MARKDOWN_MAX_LINE_LENGTH);

	/* https://spec.commonmark.org/0.30/#characters-and-lines */

	for (i=0; i<len; i++)
		if (line[i] != ' ' && line[i] != '\t') {
			ret = FALSE;
			break;
		}

	return ret;
}

/**
 * egg_markdown_to_text_line_is_blockcomment_start:
 **/
static gboolean
egg_markdown_to_text_line_is_blockcomment_start (const gchar *line) {
	return g_str_has_prefix(line, "<!--");
}

/**
 * egg_markdown_to_text_line_is_blockcomment_end:
 **/
static gboolean
egg_markdown_to_text_line_is_blockcomment_end (const gchar *line) {
	return g_strstr_len(line, -1, "-->") != 0;
}

static gboolean
egg_markdown_to_text_line_is_tag_a_name (const gchar *line) {
	gchar *p = (gchar*) line;
	gchar *q = "name";
	gchar *u = "NAME";
	gboolean ret = FALSE;

	skip_blanks (&p);
	/* want <a */
	ret = *p++ == '<' && *p && (*p == 'a' || *p == 'A');
	if (ret) {
		p++; skip_blanks (&p);
		/* want name */
		while (*p && *q && (*p == *q || *p == *u))
			p++, q++, u++;
		ret = *p && *q == '\0'; /* name matched */
		if (ret) {
			if (*p == '=') {
				q = ++p;
				ret = *q == '"' || *q == '\'';
				if (ret) { /* consume quoted value */
					p++;
					while (*p && *p != *q)
						p++;
					ret = *p == *q;
					p++;
				}
			}
		}
		if (ret) { /* want > */
			skip_blanks (&p);
			ret = *p == '>';
		}
		if (ret) { /* want </a> */
			p++; skip_blanks (&p);
			/* want </a */
			ret =	*p++ == '<' &&
				*p && *p++ == '/' &&
				*p && (*p == 'a' || *p == 'A');
			if (ret) {
				p++; skip_blanks (&p);
				ret = *p == '>';
				/* want no suffix */
				p++; skip_blanks (&p);
				ret = *p == '\0';
			}
		}
	}
	return ret;
}

/**
 * egg_markdown_to_text_line_has_line_break:
 **/
static inline gboolean
egg_markdown_to_text_line_has_line_break (const gchar *line)
{
	return g_str_has_suffix (line, "  ");
}

/**
 * egg_markdown_to_text_line_is_pipe_table_start:
 * one-line look-ahead
 **/
static gboolean
egg_markdown_to_text_line_is_pipe_table_start (const gchar *line, const gchar *next_line, guint *columns)
{
	gboolean ret = FALSE; /* default return if no table start */
	gint pipes = 0, dashes, col;
	gchar *p, *q;

	if (!*line || next_line == NULL)
		goto out;

	/* column header - count inner pipes */

	if (*(p = q = (gchar *) line) == '|')
		p++; /* don't count optional opening pipe */

	for (; *p; p++)
		if (*p == '|') {
			pipes++;
		if (*p != ' ' && *p != '\t')
			q = p;
		}

	if (*q == '|')
		pipes--; /* discount optional closing pipe */

	if (pipes <= 0)
		goto out;

	*columns = pipes + 1;

	/* header underlines - count columns */

	if (*(p = (gchar *) next_line) == '|')
		p++; /* ignore optional opening pipe */

	for (col = 0; col < *columns && *p; ++col, p++) {
		dashes = 0;
		while (*p && *p == ' ')
			p++;
		if (*p && *p == ':') /* left-align option */
			p++, dashes++;
		while (*p && *p == '-')
			p++, dashes++;
		if (*p && *p == ':') /* right-align option */
			p++, dashes++;
		while (*p && *p == ' ')
			p++;
		if (*p && *p != '|' && *p != '+')
			break;
		if (dashes < 3)
			break;
	}

	if (col < *columns)
		goto out;
	ret = TRUE;
out:
	return ret;
}

/**
 * egg_markdown_to_text_line_is_table_end:
 **/
static gboolean
egg_markdown_to_text_line_is_table_end (const gchar *line, const guint columns)
{
	gboolean ret = TRUE; /* default return if not a table row */
	gint col;
	gchar *p;
	gboolean b; /* backslash is escaped */
	gboolean e; /* previous char escapes current pipe char */

	if (line == NULL || !*(p = (gchar *) line))
		goto out;

	if (*p == '|') /* ignore optional opening pipe */
		p++;

	for (col = 0, b = e = FALSE; col < columns && *p; p++) {
		while (*p && *p == ' ')
			p++;

		/* let's stipulate that | is a column divider iff the previous
		 * char isn't \ (escape) or a character in [] */
		if (*p == '\\') {
			e = !b;
			b = !b;
		}
		/* else if (*p == '_' || *p == '*' || *p == '+') { */
		/* 	e = TRUE; */
		/* } */
		else if (*p == '|' && !e && !b)
			col++;
	}

	ret = col != columns; /* excess columns will pass */
out:
	return ret;
}

/**
 * egg_markdown_replace:
 * Replace all literal matches of @needle in @haystack with @replace
 * @haystack `gchar` array
 * @needle   `gchar` array
 * @replace  `gchar` array, replacement
 * @pcount   pointer to unsigned integer, nullable
 * Return:   the replaced string and the number of replacements in *@pcount
 **/
static gchar *
egg_markdown_replace (const gchar *haystack, const gchar *needle, const gchar *replace, guint *pcount)
{
	gchar *new;
	gchar **split;

	split = g_strsplit (haystack, needle, -1);
	if (pcount)
		*pcount = g_strv_length (split) - 1;
	new = g_strjoinv (replace, split);
	g_strfreev (split);

	return new;
}

/**
 * egg_markdown_replace_unichars:
 * Replace all matches of regex pattern @wc1(@suffix)? with @replace
 * @wc1      Unicode character
 * @suffix   `gchar` array, nullable
 * @replace  `gchar` array replacement
 * @pcount   pointer to unsigned integer, nullable
 * Return:   the replaced string and the number of replacements in *@pcount
 **/
static gchar *
egg_markdown_replace_unichars
(const gchar *haystack, const gunichar wc1, const gchar *suffix, const gchar *replace, guint *pcount)
{
	GString *s1, *s12;
	guint count = 0, n;
	gchar *ret, *p = NULL;

	s1 = g_string_new ("");
	g_string_append_unichar (s1, wc1);

	if (suffix) {
		s12 = g_string_new (s1->str);
		g_string_append (s12, suffix);
		p = egg_markdown_replace (haystack, s12->str, replace, &n);
		count += n;
		g_string_free (s12, TRUE);
	}
	ret = egg_markdown_replace (p ? p : haystack, s1->str, replace, &n);
	g_free (p);
	count += n;
	if (pcount)
		*pcount = count;
	g_string_free (s1, TRUE);

	return ret;
}

/**
 * egg_markdown_string_chomp:
 * trim suffix up to limit occurrences (pass limit -1 for all occurrences)
 **/
static inline void
egg_markdown_string_chomp (GString *s, const gchar *suffix, guint limit)
{
	g_return_if_fail (suffix && *suffix);

	for (guint d = strlen (suffix); limit && g_str_has_suffix (s->str, suffix); limit--)
		g_string_truncate (s, s->len - d);
}

#ifdef PRINTERS
/* standout, standout end */
#define _SO    "\033[7m"
#define _SE    "\033[0m"
#define _SObla "\033[7;30m"
#define _SOred "\033[7;31m"
#define _SOgre "\033[7;32m"
#define _SOyel "\033[7;33m"
#define _SOblu "\033[7;34m"
#define _SOmag "\033[7;35m"
#define _SOcya "\033[7;36m"
#define _SOwhi "\033[7;37m"
#endif

#if PRINTERS > 0
static void
_print_priv_table (const GArray *priv_table, const gchar *fmt, ...)
{
	va_list args;
	va_start (args, fmt);
	gchar *msg = g_strdup_vprintf (fmt, args);
	va_end (args);
	g_printerr ("%s", msg);
	g_free (msg);

	if (priv_table->len == 0)
		g_printerr (_SO"EMPTY TABLE"_SE"\n");
	else for (gint i = 0; i < priv_table->len; i++) {
		gchar *elem = g_array_index (priv_table, gpointer, i);
		g_printerr("% 3d. \"%s\"\n", i, elem);
	}
}

static void
_print_slices (const GArray *slices, const gchar *text, const gchar *fmt, ...)
{
	va_list args;
	va_start (args, fmt);
	gchar *msg = g_strdup_vprintf (fmt, args);
	va_end (args);
	g_printerr ("%s", msg);
	g_free (msg);

	if (slices->len == 0)
		g_printerr (_SO"NO SLICES"_SE"\n");
	else for (gint i = 0; i < slices->len; i++) {
		EggMarkdownSlice *slice = &g_array_index (slices, EggMarkdownSlice, i);
		EggMarkdownSliceType type = slice->type;
		gint start = slice->start;
		gint len   = slice->len;
		gchar data[10]; /* ~itoa */
		sprintf (data, "%d", slice->data);
		g_printerr ("% 3d. % 12s % 3d,%-3d % 3s", i, _slice_type_name[type], start, len, *data == '-' ? "" : data);
		if (text) {
			gchar *p = (gchar*)text + start;
			gchar *q = p + len;
			gchar c = *q;
			*q = '\0';
			g_printerr (" \"%s\"", p);
			*q = c;
		}
		g_printerr ("\n");
	}
}

#endif /* PRINTERS */

/**
 * egg_markdown_make_code_ref
 * @id: id of an inline code segment stashed with `egg_markdown_stash_code`.
 * Return:
 * gchar *code_ref an encoded form of @id
**/
static inline gchar*
egg_markdown_make_code_ref (const guint id) {
	return g_strdup_printf ("%s%dC%s", sUNIPUA_CODE, id, sUNIPUA_CODE);
}

/**
 * egg_markdown_get_code_id
 * @code_ref    code_ref text
 * Return: -1   if @text isn't a valid code_ref otherwise return the code id
 **/
gint
egg_markdown_get_code_id (const gchar *code_ref) {
	gint id = -1;
	GString *s = g_string_new (code_ref);
	if (g_string_replace (s, sUNIPUA_CODE, "", 1) == 1)
		id = atoi (s->str);
	g_string_free (s, TRUE);

	return id;
}

/**
 * egg_markdown_stash_code
 * @code: markdown code span or `code_ref`
 * Return:
 * stashed `code_id` - use `egg_markdown_get_code` to retrieve the code text.
 * Identical @{code}s are stashed with the same id.
 **/
static inline guint
egg_markdown_stash_code (EggMarkdown *self, const gchar *code) {
	gint id = egg_markdown_get_code_id (code);
	if (id < 0) {
		gchar *p;
		for (id = 0; id < self->priv->code_table->len; id++) {
			p = g_array_index (self->priv->code_table, gchar *, id);
			if (strcmp(code, p) == 0)
				break;
		}
		if (id == self->priv->code_table->len) {
			p = g_strdup(code);
			g_array_append_val(self->priv->code_table, p);
			id = self->priv->code_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_get_code
 * @id: id of a code span previously stashed with `egg_markdown_stash_code`.
 * Return:
 * code text with unescaped "\`"
 **/
static gchar *
egg_markdown_get_code(EggMarkdown *self, const gint id)
{
	gchar *ret = NULL;
	gboolean esc = FALSE;

	g_return_val_if_fail(id < self->priv->code_table->len, NULL);
	return g_array_index(self->priv->code_table, gchar *, id);
}

/**
 * egg_markdown_regex_code:
 * Return: GRegex* matcher for markdown code spans
 **/
static GRegex *
egg_markdown_regex_code (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/*
	* Nearly complete pattern for "https://spec.commonmark.org/0.30/#code-spans
	* This pattern can match a valid code span that extends across lines
	* (see \X*?).  However, line endings are not converted to spaces here
	* but in release_emerged_line_endings, which converts back the line
	* endings that were changed to iUNIPUA_BR while the pending line was
	* being accumulated.
	* Exceptions:
	* - ^``` is a valid start sequence for both code blocks and code spans.
	*   Commonmark assigns code spans higher priority but we do the opposite.
	*   Thus "``a\n```b```\n``" becomes
	*   <p><code>a ```b``` </code></p>    for commonmark but
	*   <p>`` a</p>
	*   <pre><code>``</code></pre>        for us.
	* - Dangling spaces in multi-line code spans are not preserved.
	*   Thus "``\nfoo \n``\n" becomes
	*   <p><code>foo </code></p>          for commonmark but
	*   <p><code>foo</code></p>           for us.
	*/

	/* see bib_regular_expression */
	/* https://regex101.com tester */
	/*	((?<!`)`+(?!`))(\x20?)(\X+?)(\g{-3})((?<!`)\g{-5}(?!`))|((?<!`)`+(?!`))(\h+)((?<!`)\g{-2}(?!`))      gm */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_CODE);
	if (*regex == NULL) {
		GError *err = NULL;
		*regex = g_regex_new (
		/* see EggMarkdownSliceType for meaning of UPPERCASE words */
		/* leading group number is the switch{} case label         */
		/* capture group numbers are used in egg_markdown_slice_get_codes switch{} */
			"((?<!`)`+(?!`))"              /* 1 MID  */
			"(\\x20?)"                     /* 2 TICK */
			"(\\X+?)"                      /* 3 CODE */
			"(\\g{-3})"                    /* 4 TICK */
			"((?<!`)\\g{-5}(?!`))"         /* 5 END  */
			"|((?<!`)`+(?!`))"             /* 6 MID  */
			"(\\h+)"                       /* 7 CODE */
			"((?<!`)\\g{-2}(?!`))"         /* 8 END  */
			, 0, 0, &err);
	/*
	* https://spec.commonmark.org/0.30/#example-341
	* Code span backticks have higher precedence than any other inline
	* constructs except HTML tags and autolinks. */
	}

	return *regex;
}

/**
 * egg_markdown_regex_code_ref:
 * Return: GRegex* matcher for internal code_refs
 **/
static GRegex *
egg_markdown_regex_code_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* code_ref created by egg_markdown_make_code_ref */
	/* and normalized by egg_markdown_slice_join_codes */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_CODE_REF);
	if (*regex == NULL)
		*regex = g_regex_new ("`"sUNIPUA_CODE"(\\d+)C"sUNIPUA_CODE"`" , 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_slice_find_type
 * @slices GArray of @EggMarkdownSlice elements obtained with
 * egg_markdown_slice_get_codes or egg_markdown_slice_get_links.
 * @type:   @EggMarkdownSliceType value to find in @slices
 * @start:  array index where the search starts.
 * @retptr: address of an @EggMarkdownSlice pointer, NULLable.
 * Return:
 * `gint` array index of the first slice of type @type found;
 * a pointer to the value is returned in *@retptr.
 * -1 if no match, and *@retprt is undefined.
 **/
static gint
egg_markdown_slice_find_type (const GArray *slices, const EggMarkdownSliceType type, const gint start, EggMarkdownSlice **retptr)
{
	gint ret = -1, i;
	EggMarkdownSlice *slice;
	for (i = start; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		if (slice->type == type) {
			if (retptr)
				*retptr = slice;
			ret = i;
			break;
		}
	}
	return (ret);
}

/**
 * egg_markdown_slice_get_codes:
 * @text: the markdown text to harvest for inline code slices.
 * @stash_codes:  request stashing inline code texts.
 * Return:
 * `GArray` of @EggMarkdownSlice elements.
 * Array length equals to zero means no match.
 *
 * @text is sliced to facilitate extracting the text of markdown code spans
 * like `CODE`.  A slice value of type EGG_MARKDOWN_SLICE_CODE_TYPE_TEXT
 * includes the start and end positions of the CODE component.  Joining all
 * slices together yields the original @text.
 *
 * If @stash_codes is TRUE each CODE is stashed into @self's code_table.
 **/
static GArray *
egg_markdown_slice_get_codes (EggMarkdown *self, const gchar *text, const gboolean stash_codes)
{
	GRegex *regex = egg_markdown_regex_code (self);
	GMatchInfo *match_info = NULL;
	gint i, END, start = 0, end = 0, done_end = 0;
	EggMarkdownSlice slice;
	GArray *slices;

	slices = g_array_new (FALSE, FALSE, sizeof (slice));
	g_regex_match (regex, text, 0, &match_info);
	while (g_match_info_matches (match_info)) {
		END = g_regex_get_capture_count (regex);
		for (i = 0; i <= END; i++, start = end = 0) {
			(void)g_match_info_fetch_pos (match_info, i, &start, &end);
			if (start < 0)
				continue;
			slice.start = start;
			slice.len   = end - start;
			slice.data  = -1;
			slice.type  = EGG_MARKDOWN_SLICE_UNKNOWN;
			/* switch case label is the capture group number in egg_markdown_regex_code */
			switch(i) {
				case 0:
					slice.type = EGG_MARKDOWN_SLICE_BEFORE;
					slice.start = done_end;
					slice.len = start - done_end;
					done_end = end - 1;
					g_array_append_val (slices, slice);

					slice.type = EGG_MARKDOWN_SLICE_MATCH;
					slice.start = start;
					slice.len   = end - start;
					break;
				case 1:
				case 6:
					slice.type = EGG_MARKDOWN_SLICE_CODE_MID;
					break;
				case 2:
				case 4:
					if (slice.len == 0) /* omit empty TICKs */
						continue;
					slice.type = EGG_MARKDOWN_SLICE_CODE_TICK;
					break;
				case 3:
				case 7:
					slice.type = EGG_MARKDOWN_SLICE_CODE_CODE;
					if (stash_codes) {
						gchar *code;
						code = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_code (self, code);
						g_free (code);
					}
					break;
				case 5:
				case 8:
					slice.type = EGG_MARKDOWN_SLICE_CODE_END;
					done_end = slice.start + slice.len;
					break;

			}
			g_array_append_val (slices, slice);
		}
		g_match_info_next (match_info, NULL);
	}

	if (slices->len > 0) {
		/* capture the text after the last END */
		slice.type  = EGG_MARKDOWN_SLICE_AFTER;
		slice.start = done_end;
		slice.len   = strlen (text) - done_end;
		slice.data  = -1;
		g_array_append_val (slices, slice);
#if PRINTERS > 0
		_print_slices (slices, text,
			_SOgre"%s: CODE SLICES:"_SE" \"%s\"\n", __FUNCTION__, text);
#endif
	}
	g_match_info_free (match_info);

	return (slices);
}

/**
 * egg_markdown_make_link_dest_ref
 * @id: id of a link DEST previously stashed with `egg_markdown_stash_link_dest`.
 * Return:
 * `link_dest_ref`: encoded @id
 *
 * This function also applies to markdown image path.
**/
static inline gchar*
egg_markdown_make_link_dest_ref (const guint id) {
	return g_strdup_printf ("%s%dLu%s", sUNIPUA_LINK, id, sUNIPUA_LINK);
}

/**
 * egg_markdown_make_link_title_ref
 * @id: id of a link title previously stashed with `egg_markdown_stash_link_title`.
 * Return:
 * `link_title_ref`: encode @id
**/
static inline gchar*
egg_markdown_make_link_title_ref (const guint id) {
	return g_strdup_printf ("%s%dLt%s", sUNIPUA_LINK, id, sUNIPUA_LINK);
}

/**
 * egg_markdown_get_link_dest_id
 * @ref    a `link_dest_ref`
 * Return: -1   if @ref isn't a valid link_dest_ref otherwise return the link DEST id
 **/
static gint
egg_markdown_get_link_dest_id (const gchar *ref) {
	gint id = -1;
	GString *s = g_string_new (ref);
	if (g_string_replace (s, sUNIPUA_LINK, "", 1) == 1)
		id = atoi (s->str);
	g_string_free (s, TRUE);

	return id;
}

/**
 * egg_markdown_get_link_title_id
 * @ref    a `link_title_ref`
 * Return: -1   if @ref isn't a valid link_title_ref otherwise return the link title id
 **/
static gint
egg_markdown_get_link_title_id (const gchar *ref) {
	return egg_markdown_get_link_dest_id (ref);
}

/**
 * egg_markdown_stash_link_dest
 * @dest: markdown link destination or `link_dest_ref`
 * Return:
 * stashed `link_dest_id` - use `egg_markdown_get_link_dest` to retrieve the link destination.
 * Identical destinations are stashed with the same id.
 *
 * This function also applies to markdown image path.
 **/
static gint
egg_markdown_stash_link_dest (EggMarkdown *self, const gchar *dest) {
	gint id = egg_markdown_get_link_dest_id (dest);
	if (id < 0) {
		gchar *p;
		for (id = 0; id < self->priv->link_table->len; id++) {
			p = g_array_index (self->priv->link_table, gchar *, id);
			if (strcmp(dest, p) == 0)
				break;
		}
		if (id == self->priv->link_table->len) {
			p = g_strdup(dest);
			g_array_append_val(self->priv->link_table, p);
			id = self->priv->link_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_stash_link_title
 * @title: markdown link title
 * Return:
 * stashed `link_title_id` - use `egg_markdown_get_link_title` to retrieve the title
 * Identical @{title}s are stashed with the same id.
 **/
static gint
egg_markdown_stash_link_title (EggMarkdown *self, const gchar *title) {
	gint id = egg_markdown_get_link_title_id (title);
	if (id < 0) {
		gchar *p;
		for (id = 0; id < self->priv->link_table->len; id++) {
			p = g_array_index (self->priv->link_table, gchar *, id);
			if (strcmp(title, p) == 0)
				break;
		}
		if (id == self->priv->link_table->len) {
			gchar *p = g_strdup (title);
			g_array_append_val(self->priv->link_table, p);
			id = self->priv->link_table->len - 1;
		}
	}

	return id;
}

/**
 * egg_markdown_get_link_dest
 * @id: link destination id in @self's link_table
 * This function also returns link URI or image path.
 **/
gchar *
egg_markdown_get_link_dest(EggMarkdown *self, const gint id)
{
	g_return_val_if_fail(id < self->priv->link_table->len, NULL);
	return g_strdup(g_array_index(self->priv->link_table, gchar *, id));
}

/**
 * egg_markdown_get_link_title
 * @id: link title in @self's link_table
 **/
static gchar *
egg_markdown_get_link_title(EggMarkdown *self, const gint id)
{
	return egg_markdown_get_link_dest (self, id);
}

/**
 * egg_markdown_regex_link:
 * Return: GRegex* matcher for markdown link and image syntax
 **/
static GRegex *
egg_markdown_regex_link (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/*
	* Nearly complete pattern for https://spec.commonmark.org/0.30/#links
	* and https://spec.commonmark.org/0.30/#images
	* Exceptions:
	* - DEST  : https://spec.commonmark.org/0.30/#link-destination
	* - - pointy brackets "(<...>)" are not supported
	* - TITLE : https://spec.commonmark.org/0.30/#link-title
	* - - balanced parenthesis "(...)" not implemented
	* - - escaped quotes not implemented
	* - smart quotes as delimiters : they are not in the spec but I had to add
	*   them to work around mdview's smart_quoting mode
	*/

bib_regular_expression:
	/* https://docs.gtk.org/glib/struct.Regex.html */
	/* https://perldoc.perl.org/perlre */
	/* https://www.regular-expressions.info/unicode.html */

	/* https://regex101.com tester */
	/*	(?<TYPE>\!?\[)(?<TEXT>[^]]+?)?(\]\x28\h*)(?<DEST>[^\x29]+?)?(?:(\h+([“‘]|(['\"])))(?<TITLE>.+?)?([’”]|\g{-3}))?(\h*\x29)   gm */

	/* TODO
	* Considering that code spans have higher precedence than links,
	* do I need to add sUNIPUA_CODE as a DEST match stopper?
	* https://spec.commonmark.org/0.30/#example-341 */
	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK);
	if (*regex == NULL) {
		GError *err = NULL;
		*regex = g_regex_new (
		/* see EggMarkdownSliceType for meaning of UPPERCASE words */
		/* capture group numbers are used in egg_markdown_slice_get_links switch{} */
		/* capture group names are used in _formatter_link_cb */
			"(?<TYPE>\\!?\\[)"           /* 1 TYPE   */
			"(?<TEXT>[^]]+?)?"           /* 2 TEXT   */
			"(\\]\\x28\\h*)"             /* 3 MID (  */
			"(?<DEST>[^\\x29]+?)?"       /* 4 DEST ) */
		"(?:"
			"(\\h+"                      /* 5 MID, open quote */
				"([“‘]|"                   /* 6 smart/straight quotes, ignore in switch{} */
				"(['\"]))"                 /* 7 straight quote, ignore in switch{} */
			")"
			"(?<TITLE>.+?)?"             /* 8 TITLE  */
			"([’”]|\\g{-3})"             /* 9 END, close quote, merge next ... */
		")?"
			"(\\h*\\x29)"                /* 10 ... END, merge previous  */
		, 0, 0, &err);
	/*
	* https://spec.commonmark.org/0.30/#example-341
	* Code span backticks have higher precedence than any other inline
	* constructs except HTML tags and autolinks. */
	}

	return *regex;
}

/**
 * egg_markdown_regex_link_dest_ref:
 * Return: GRegex* matcher for internal link_dest_refs
 **/
static GRegex *
egg_markdown_regex_link_dest_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* link_dest_refs created by egg_markdown_make_link_dest_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK_DEST_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_LINK"(\\d+)Lu"sUNIPUA_LINK , 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_regex_link_title_ref:
 * Return: GRegex* matcher for internal link_title_refs
 **/
static GRegex *
egg_markdown_regex_link_title_ref (EggMarkdown *self)
{
	static GRegex **regex = NULL;

	/* link_title_refs created by egg_markdown_make_link_title_ref */

	if (regex == NULL)
		regex = &g_array_index(self->priv->regex_table, GRegex *, EGG_MARKDOWN_REGEX_LINK_TITLE_REF);
	if (*regex == NULL)
		*regex = g_regex_new (sUNIPUA_LINK"(\\d+)Lt"sUNIPUA_LINK , 0, 0, NULL);

	return *regex;
}

/**
 * egg_markdown_slice_get_links:
 * @text: the markdown text to harvest for link slices.
 * @stash_links:  request stashing links.
 * Return: GArray of EggMarkdownSlice values.
 *
 * @text is sliced to facilitate extracting the components of link and image
 * markdown like `[text](LINK)` and `![alt text](LINK)`.  A slice value of type
 * EGG_MARKDOWN_SLICE_LINK_TYPE_URL or EGG_MARKDOWN_SLICE_LINK_TYPE_DEST
 * includes the start and end positions of the LINK component.  Joining all
 * slices together yields the original @text.
 *
 * If @stash_links is TRUE each LINK is stashed into @self's link_table.
 **/
static GArray *
egg_markdown_slice_get_links (EggMarkdown *self, const gchar *text, const gboolean stash_links)
{

	GRegex *regex = egg_markdown_regex_link(self);
	GMatchInfo *match_info = NULL;
	gint i, END, start = 0, end = 0, done_end = 0;
	EggMarkdownSlice slice;
	GArray *slices;


	slices = g_array_new (FALSE, FALSE, sizeof (slice));
	g_regex_match (regex, text, 0, &match_info);
	while (g_match_info_matches (match_info)) {
		END = g_regex_get_capture_count (regex);
		for (i = 0; i <= END; i++, start = end = 0) {
			(void)g_match_info_fetch_pos (match_info, i, &start, &end);
			if (start < 0)
				continue;
			slice.start = start;
			slice.len   = end - start;
			slice.data  = -1;
			slice.type  = EGG_MARKDOWN_SLICE_UNKNOWN;
			/* switch case label is the capture group number in egg_markdown_regex_link */
			switch(i) {
				case 0:
					slice.type = EGG_MARKDOWN_SLICE_BEFORE;
					slice.start = done_end;
					slice.len = start - done_end;
					done_end = end - 1;
					g_array_append_val (slices, slice);

					slice.type = EGG_MARKDOWN_SLICE_MATCH;
					slice.start = start;
					slice.len   = end - start;
					break;
				case 3:
				case 5:
					slice.type = EGG_MARKDOWN_SLICE_LINK_MID;
					break;
				case 6:
				case 7:
					slice.type = EGG_MARKDOWN_SLICE_IGNORED;
					continue;
				case 1:
					switch (text[start]) {
						case '!': slice.type = EGG_MARKDOWN_SLICE_LINK_TYPE_IMG; break;
						case '[': slice.type = EGG_MARKDOWN_SLICE_LINK_TYPE_URL; break;
					}
					break;
				case 2:
					slice.type = EGG_MARKDOWN_SLICE_LINK_TEXT;
					break;
				case 4:
					slice.type = EGG_MARKDOWN_SLICE_LINK_DEST;
					if (stash_links && slice.len > 0) {
						gchar *link = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_link_dest (self, link);
						g_free (link);
					}
					break;
				case 8:
					slice.type = EGG_MARKDOWN_SLICE_LINK_TITLE;
					if (stash_links) {
						gchar *title = g_strndup (text + slice.start, slice.len);
						slice.data = egg_markdown_stash_link_title (self, title);
						g_free (title);
					}
					break;
				case 9:
					/* start the end segment */
					done_end = slice.start;
					/* merge into next */
					continue;
				case 10:
					slice.type = EGG_MARKDOWN_SLICE_LINK_END;
					/* merge previous */
					slice.len += slice.start - done_end;
					slice.start = done_end;
					done_end = slice.start + slice.len;
					break;
			}
			g_array_append_val (slices, slice);
		}
		g_match_info_next (match_info, NULL);
	}

	if (slices->len > 0) {
		/* capture the text after the last END */
		slice.type  = EGG_MARKDOWN_SLICE_AFTER;
		slice.start = done_end;
		slice.len   = strlen (text) - done_end;
		slice.data  = -1;
		g_array_append_val (slices, slice);
#if PRINTERS > 0
		_print_slices (slices, text,
			_SOgre"%s: LINK SLICES:"_SE" %s\n", __FUNCTION__, text);
#endif
	}
	g_match_info_free (match_info);

	return (slices);
}

/**
 * egg_markdown_slice_join_links
 * @slices  @text's link slices from egg_markdown_slice_get_links
 * @text    markdown text
 * Return:  joined markdown text with link_dest_ref and link_title_ref replacements
 **/
static gchar *
egg_markdown_slice_join_links (const GArray *slices, const gchar* text)
{
	EggMarkdownSlice *slice;
	gint done_end = 0, i;
	gchar *yp;
	GString *buf = g_string_new ("");

	for (i = 0; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		switch (slice->type) {
			case EGG_MARKDOWN_SLICE_LINK_DEST:
				if (slice->start < 0)
					break;
				/* all non-link segments before the link */
				g_string_append_len (buf, text + done_end, slice->start - done_end);

				/* replace link_dest_ref for DEST */
				yp = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_link_dest_id (yp) < 0) { /* yp not yet a link_dest_ref */
					gchar *ref = egg_markdown_make_link_dest_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else {
					g_string_append (buf, yp);
				}
				g_free (yp);

				done_end = slice->start + slice->len;
				break;
			case EGG_MARKDOWN_SLICE_LINK_TITLE:
				if (slice->start < 0)
					break;
				g_string_append_len (buf, text + done_end, slice->start - done_end);

				/* replace link_title_ref for TITLE */
				yp = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_link_title_id (yp) < 0) { /* yp not yet a link_title_ref */
					gchar *ref = egg_markdown_make_link_title_ref (slice->data);
					g_string_append (buf, ref);
					g_free (ref);
				} else {
					g_string_append (buf, yp);
				}
				g_free (yp);

				done_end = slice->start + slice->len;
				break;
		}
	}
	/* non-link text tail */
	g_string_append (buf, text + done_end);

	return g_string_free (buf, FALSE);
}

static gchar *egg_markdown_word_auto_format_code_segment (const gchar *text);
/* helper for egg_markdown_slice_join_codes */
static void
_append_autocode (GString *buf, gchar *nc, const guint len)
{
	gint b, a; /* (b)efore / (a)fter the non-code span nc */
	gchar *sb = NULL, *sa = NULL;
#if PRINTERS > 0
	g_print ("auto-code for \"%s\"", nc);
#endif
	/* don't want to start/end the non-code span with backticks */
	for (b = 0; nc[b] && nc[b] == '`'; b++)
		;
	for (a = 0; len > b && nc[a] && nc[a] == '`'; a++)
		;
	sb = g_strnfill (b, '`');
	sa = g_strnfill (a, '`');
	if (a) nc[len - a] = '\0';

	gchar *ac = egg_markdown_word_auto_format_code_segment (b ? nc + b : nc);
#if PRINTERS > 0
	g_print (" ==> %s\"%s\"\n", ac ? "" : "(no change) ", ac ? ac : nc);
#endif
	g_string_append_printf (buf, "%s%s%s", sb, ac ? ac : nc, sa);
	g_free (ac);
	g_free (sb);
	g_free (sa);
	if (a) nc[len - a] = '`';
}

/**
 * egg_markdown_slice_join_codes
 * @slices    @text's code slices from egg_markdown_slice_get_codes
 * @text      markdown text
 * @insert_autocodes  if TRUE then assess and replace auto-codes in non-code segments
 * Return:  joined markdown text with code_ref replacements and auto-codes
 **/
static gchar *
egg_markdown_slice_join_codes (const GArray *slices, const gchar* text, const gboolean insert_autocodes)
{
	g_warn_if_fail (slices);
	g_warn_if_fail (text);
	EggMarkdownSlice *slice;
	gint done_end = 0, i, len, match_start;
	gchar *yc = NULL, *nc = NULL;
	GString *buf = g_string_new ("");

#if PRINTERS > 0
	if (insert_autocodes)
		g_printerr (_SOyel"%s: AUTOCODE DISCOVERY:"_SE"\n", __FUNCTION__);
#endif
	/*
	 * Sequence: START ( MID [TICK] CODE [TICK] )* END
	 * Normalization: replace inside brackets above with `code_ref`
	 * IMPORTANT: exactly 1 backtick before and 1 after code_ref and no TICKs
	*/
	gboolean skip_to_END = FALSE;
	for (i = 0; i < slices->len; i++) {
		slice = &g_array_index (slices, EggMarkdownSlice, i);
		if (slice->start < 0)
			continue;
		switch (slice->type) {
			case EGG_MARKDOWN_SLICE_BEFORE:
				match_start = slice->start;
				break;
			case EGG_MARKDOWN_SLICE_CODE_MID:
				if (skip_to_END)
					break;

				/* the match start game */
				if (match_start >= 0) {
					len = slice->start - match_start;
					if (len > 0) {
						nc = g_strndup (text + match_start, len);
						if (insert_autocodes)
							_append_autocode (buf, nc, len);
						else
							g_string_append (buf, nc);
						g_free (nc);
						done_end = done_end + len;
					}
					match_start = -1;
				}

				len = slice->start - done_end;
				/* flush the non-code (nc) segments before MID */
				if (len > 0) {
					nc = g_strndup (text + done_end, len);
					if (insert_autocodes)
						_append_autocode (buf, nc, len);
					else
						g_string_append (buf, nc);
					g_free (nc);
				}

				/* go on until the corresponding CODE is found */
				break;

			case EGG_MARKDOWN_SLICE_CODE_CODE:
				g_string_append (buf, "`");

				/* replace code_ref for CODE */
				yc = g_strndup (text + slice->start, slice->len);
				if (slice->data >= 0 && egg_markdown_get_code_id (yc) < 0) { /* yc isn't a code_ref */
					gchar *code_ref = egg_markdown_make_code_ref (slice->data);
					g_string_append (buf, code_ref);
					g_free (code_ref);
				} else
					g_string_append (buf, yc);
				g_free (yc);

				g_string_append (buf, "`");

				skip_to_END = TRUE;
				break;

			case EGG_MARKDOWN_SLICE_CODE_END:
				skip_to_END = FALSE;
				done_end = slice->start + slice->len;
				break;
		}
	}

	/* the end game: text's non-code (nc) tail, if any */
	len = strlen (text);
	if (done_end < len) {
		if (slices->len > 0) {
			nc = g_strdup(text + done_end);
			if (insert_autocodes)
				_append_autocode (buf, nc, len - done_end);
			else
				g_string_append_len (buf, nc, len - done_end);
			g_free (nc);
		} else {
			g_string_assign (buf, text);
		}
	}

	return g_string_free (buf, FALSE);
}

/**
 * egg_markdown_strstr_spaces:
 **/
static gchar *
egg_markdown_strstr_spaces (const gchar *haystack, const gchar *needle)
{
	gchar *found;
	const gchar *haystack_new = haystack;

retry:
	/* don't find if surrounded by spaces or preceeded by backslash */
	found = strstr (haystack_new, needle);
	if (found == NULL)
		return NULL;

	/* start of the string, always valid */
	if (found == haystack)
		return found;

	/* end of the string, always valid */
	if (( *(found-1) == ' ' && *(found+1) == ' ') ||
	      *(found-1) == '\\' )
	{
		haystack_new = found+1;
		goto retry;
	}
	return found;
}

/**
 * egg_markdown_to_text_line_formatter:
 **/
static gchar *
egg_markdown_to_text_line_formatter (const gchar *line, const gchar *formatter, const gchar *left, const gchar *right)
{
	guint len;
	gchar *str1;
	gchar *str2;
	gchar *start = NULL;
	gchar *middle = NULL;
	gchar *end = NULL;
	gchar *copy = NULL;
	gchar *data = NULL;
	gchar *temp;

	/* needed to know for shifts */
	len = strnlen (formatter, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* find sections */
	copy = g_strdup (line);
	str1 = egg_markdown_strstr_spaces (copy, formatter);
	if (str1 != NULL) {
		*str1 = '\0';
		str2 = egg_markdown_strstr_spaces (str1+len, formatter);
		if (str2 != NULL) {
			*str2 = '\0';
			middle = str1 + len;
			start = copy;
			end = str2 + len;
		}
	}

	/* if we found, replace and keep looking for the same string */
	if (start != NULL && middle != NULL && end != NULL) {
		temp = g_strdup_printf ("%s%s%s%s%s", start, left, middle, right, end);
		/* recursive */
		data = egg_markdown_to_text_line_formatter (temp, formatter, left, right);
		g_free (temp);
	} else {
		/* not found, keep return as-is */
		data = g_strdup (line);
	}
out:
	g_free (copy);
	return data;
}

/**
 * _resolve_ref:
 * Helper for _formatter_link_cb.
 * Resolve a link_dest_ref or link_title_ref, if any, fallback on NULL
 * @text `gchar *` link destination or link title or link_dest_ref or link_title_ref.
 * @get_regex `GRegex *` function returning the regex that can match a reference in @text.
 * @get_ref `gchar *` function returning the value of a matched reference from its id.
 * @ptrid `gint *` to store the id of the matched reference, nullable.
 * Return:
 * a newly-allocated `gchar *` containing the resolved value otherwise NULL if
 * *@get_regex did not match or there is no value for the matched reference;
 * *@prtid, if not NULL, gets the resolved reference id.
 **/
static gchar *
_resolve_ref (const EggMarkdown *self, const gchar *text, GRegex *get_regex(), gchar *get_ref(), gint *ptrid)
{
	GMatchInfo *info = NULL;
	gchar *ret = NULL, *match = NULL;
	gint start, id = -1;
	GRegex *regex = get_regex (self);

	if (g_regex_match (regex, text, 0, &info)) {
		if (g_match_info_fetch_pos (info, 1, &start, NULL) && start >= 0) {
			id = atoi (text + start);
			ret = get_ref (self, id);
		}
	}
	g_match_info_free (info);
	if (ptrid)
		*ptrid = id;
	return ret;
}

/**
 * _formatter_link_cb:
 * Callback to resolve a link_dest_ref and a link_title_ref, if any, and format
 * a link or an image (based on its TYPE match) for the current output surface.
 * This is used for markdown links and images.
 **/
static gboolean
_formatter_link_cb (const GMatchInfo *match_info, GString *buf, gpointer data)
{
	EggMarkdown *self = data;
	static gint TYPE = -2, TEXT = -2, DEST = -2, TITLE = -2;
	gchar *type = NULL, *text = NULL, *dest = NULL, *title = NULL;
	gchar *formatted = NULL, *p;
	GRegex *regex;
	GMatchInfo *info;
	gint id = -1, link_dest_id = -1, start = -1;

	/* get constant group number corresponding to named group capture */
	if (TYPE == -2) {
		regex = egg_markdown_regex_link (self);
		TYPE  = g_regex_get_string_number (regex, "TYPE");
		TEXT  = g_regex_get_string_number (regex, "TEXT");
		DEST  = g_regex_get_string_number (regex, "DEST");
		TITLE = g_regex_get_string_number (regex, "TITLE");
	}

	/* get matched captures, NULL if not captured */
	if (g_match_info_fetch_pos (match_info, TYPE, &start, NULL) && start >= 0)
		type  = g_match_info_fetch (match_info, TYPE);
	if (g_match_info_fetch_pos (match_info, TEXT, &start, NULL) && start >= 0)
		text  = g_match_info_fetch (match_info, TEXT);
	if (g_match_info_fetch_pos (match_info, DEST, &start, NULL) && start >= 0)
		dest  = g_match_info_fetch (match_info, DEST);
	if (g_match_info_fetch_pos (match_info, TITLE, &start, NULL) && start >= 0)
		title = g_match_info_fetch (match_info, TITLE);

	/* resolve link_dest_ref, if any, else fallback on dest as is, including NULL */
	if (dest) {
		dest = _resolve_ref (self, p = dest, egg_markdown_regex_link_dest_ref,
			egg_markdown_get_link_dest, &id);
		link_dest_id = id;
		g_free (p);
	}

	/* resolve link_title_ref, if any, fallback on title as is, including NULL */
	if (title) {
		title = _resolve_ref (self, p = title, egg_markdown_regex_link_title_ref,
			egg_markdown_get_link_title, &id);
			g_free (p);
	}

	/* finally format the replacement based on TYPE */
	if (type && *type == '[') { /* link */
		formatted = self->priv->tags.link_builder(text, dest, title, link_dest_id);
		g_string_append (buf, formatted);
	}
	else if (type && *type == '!') { /* image */
		formatted = self->priv->tags.image_builder(text, dest, title, link_dest_id);
		g_string_append (buf, formatted);
	}
	else
		/* shouldn't happen */
		g_printerr ("%s: internal error\n", __FUNCTION__);

	g_free (formatted);
	g_free (type);
	g_free (text);
	g_free (dest);
	g_free (title);

	return FALSE;
}

/**
 * egg_markdown_to_text_line_formatter_link:
 * Resolve link_dest_refs and link_title_refs, if any, and format link or image
 * for the current output surface.
 **/
static gchar *
egg_markdown_to_text_line_formatter_link (EggMarkdown *self, const gchar *line)
{
	GRegex *regex;
	GError *err = NULL;

	regex = egg_markdown_regex_link (self);
	gchar *ret = g_regex_replace_eval (regex, line, -1, 0, 0, _formatter_link_cb, self, &err);
	if (err) {
		g_printerr ("%s internal error: %s\n", __FUNCTION__, err->message);
		g_error_free (err);
		ret = g_strdup (line);
	}
#if defined PRINTERS
	g_printerr (_SO"%s:"_SE"\t\"%s\"\n", __FUNCTION__, ret);
#endif
	return ret;
}

/**
 * _formatter_code_span_cb:
 * Callback to resolve a code_ref match, if any, and format the code span for
 * the current output surface.
 **/
static gboolean
_formatter_code_span_cb (const GMatchInfo *match_info, GString *buf, gpointer data)
{
	EggMarkdown *self = data;
	gchar *match = g_match_info_fetch (match_info, 1);
	gboolean ret = TRUE;
	gint id = atoi (match);
	gchar *code = egg_markdown_get_code (self, id);
	if (code) {
		gchar *escaped = g_markup_escape_text(code, -1);
		g_string_append_printf (buf, "%s%s%s", self->priv->tags.code_start, escaped, self->priv->tags.code_end);
		g_free (escaped);
		ret = FALSE;
	}
	g_free (match);

	return ret;
}

/**
 * egg_markdown_to_text_line_formatter_code:
 * Resolve code_refs, if any, and format code span for the current output
 * surface.
 **/
static gchar *
egg_markdown_to_text_line_formatter_code (EggMarkdown *self, const gchar *line)
{
	GRegex *regex = egg_markdown_regex_code_ref (self);
	GError *err = NULL;

	gchar *ret = g_regex_replace_eval (regex, line, -1, 0, 0, _formatter_code_span_cb, self, &err);
	if (err) {
		g_printerr ("%s internal error:\t%s\n", __FUNCTION__, err->message);
		g_error_free (err);
		ret = g_strdup (line);
	}
#if defined PRINTERS
	g_printerr (_SO"%s:"_SE"\t\"%s\"\n", __FUNCTION__, ret);
#endif
	return ret;
}

/**
 * egg_markdown_to_text_line_formatter_exec:
	* @line: input text with PUA characters
 **/
static gchar *
egg_markdown_to_text_line_formatter_exec (EggMarkdown *self, const gchar *line)
{
	gchar *str1;
	gchar *str2;
	gchar *start = NULL;
	gchar *cmd = NULL;
	gchar *end = NULL;
	gchar *copy = NULL;
	gchar *data = NULL;

	/* @line contains PENDING_SEPARATOR.
	* @line contains iUNIPUA_BR, which does not affect command parsing because
	* it is outside command sections. However, command shall output a markdown
	* hard line-break, if any, as iUNIPUA_BR instead of \x20\x20$
	*/

	// a silly state-machine to "parse" quoting
	#define NQ  0
	#define SQ  1
	#define DQ  2
	int next[3][3] = { // current state, next state
		{NQ, SQ, DQ},  // no-quote with input of no-quote, single, double-quotes
		{SQ, NQ, SQ},  // single-quote with input of no-quote, single, double-quotes
		{DQ, DQ, NQ}   // double-quote with input of no-quote, single, double-quotes
	};
	int quote_allowed[3] = { // state
		1, 0, 1  // quote only allowed during no-quote and double-quotes
	};
	int state; // as #defined above: 0=no quote, 1=in-single-quote, 2=in-double-quote
	int next_token_is_never_quote;
	int tok;

	/* find sections */
	copy = g_strdup (line);
	str1 = egg_markdown_strstr_spaces (copy, "$(");
	if (str1 != NULL) {
		*str1 = '\0';
		state = next_token_is_never_quote = 0;
		for (str2=str1+2; *str2; str2++) {
			tok = NQ;
			if (!next_token_is_never_quote) {
				switch (*str2) {
					case ')' : if (state==NQ) goto sm_out; else break;
					case '\'' : tok=SQ; break;
					case '\"' : tok=DQ; break;
					case '\\' : next_token_is_never_quote = quote_allowed[state];
					break;
				}
			} else next_token_is_never_quote=0; //reset for next
			state = next[state][tok];
		}
sm_out:
		if (state==NQ) {
			*str2 = '\0';
			start = copy;
			cmd = str1 + 2;
			end = str2 + 1;
		}
	}

	/* if we found, replace and keep looking for the same string */
	if (start && (cmd && *cmd) && end) {
		GString *temp = g_string_sized_new(EGG_MARKDOWN_MAX_LINE_LENGTH);
		g_string_assign (temp, start);

		FILE *p = popen(cmd,"r");
		char *line = g_malloc0 (EGG_MARKDOWN_MAX_LINE_LENGTH);
		int linelen;
		while (fgets(line, EGG_MARKDOWN_MAX_LINE_LENGTH, p)) {
			linelen = strlen(line);
			if (linelen > 0) {
				if (line[linelen-1]=='\n') line[linelen-1]=0;
				g_string_append (temp, line);
			}
		}
		pclose(p);
		g_free (line);

		g_string_append (temp, end);
		data = egg_markdown_to_text_line_formatter_exec (self, temp->str);
		g_string_free (temp, TRUE);
	} else {
		/* not found, keep return as-is */
		data = g_strdup (line);
	}

	g_free (copy);
	return data;
}

static gchar *
egg_markdown_escape_c_str(const gchar *line)
{
	gchar c;
	GString *s = g_string_sized_new (strlen(line)*2);

	for (; (c = *line); line++) {
		switch(c) {
			case '"':
			case '\\':
				g_string_append_c (s,'\\');
			default:
				g_string_append_c (s,c);
		}
	}

	return g_string_free (s, FALSE);
}

/**
 * egg_markdown_to_text_line_formatter_gettext:
	* @line: input text with PUA characters
 **/
static gchar *
egg_markdown_to_text_line_formatter_gettext (EggMarkdown *self, const gchar *line)
{
	/* Considering translators.
	 * Input text contains PENDING_SEPARATOR, which would look quite awkward
	 * if included in a gettext MSGID.  @line contains the iUNIPUA_BR
	 * Unicode PUA character, which could be invisible or pose issues to
	 * translators or their tools.
	 * Therefore return MSGID replacing MSGSTR_BR for iUNIPUA_BR. */
	/* Translators shall to translate MSGSTR_BR. */
	/* @line shall not not contain MSGSTR_BR. */

	if (self->priv->pot_output)
	{
		if (!self->priv->temporary_disable_pot) {
			gchar *escaped = egg_markdown_escape_c_str(line);

			/* replace iUNIPUA_BR => MSGSTR_BR */
			gchar *msgid = egg_markdown_replace_unichars (escaped,
					iUNIPUA_BR, NULL,
					MSGSTR_BR, NULL);
			g_free (escaped);

			fprintf(self->priv->pot_output,
			"#: %s:%u\n"
			"msgid \"%s\"\nmsgstr \"\"\n\n",
			self->priv->input_filename, self->priv->curr_line_idx + 1, msgid);
			g_free (msgid);
			return g_strdup(line);
		} else {
			return g_strdup("");
		}
	} else if (self->priv->use_gettext)
	{
		if (!self->priv->temporary_disable_pot) {
			gchar *ret, *msgid, *msgstr;
			GString *br;

			msgid = egg_markdown_replace_unichars (line,
				iUNIPUA_BR, NULL,
				MSGSTR_BR, NULL);

			msgstr = g_strdup(gettext(msgid));
			g_free (msgid);

			br = g_string_new ("");
			g_string_insert_unichar (br, -1, iUNIPUA_BR);
			ret = egg_markdown_replace (msgstr, MSGSTR_BR,
					br->str, NULL);
			g_string_free (br, TRUE);
			g_free (msgstr);
			return ret;
		} else {
			return g_strdup(line);
		}
	}
}

/**
 * egg_markdown_to_text_line_formatter_remove_backslash:
 **/
static gchar *
egg_markdown_to_text_line_formatter_remove_backslash (gchar *line)
{
	gchar c;
	gboolean quoted;
	GString *s = g_string_sized_new (strlen(line)*2);

	for (quoted=FALSE; (c = *line); line++) {
		if (quoted) {
			switch(c) {
				case '\\': case '`':
				case '*': case '_' :
				case '{': case '}':
				case '[': case ']':
				case '(': case ')':
				case '#':
				case '+': case '-':
				case '.': case '!':
					// remove backslash
					break;
				default:
					// all other - insert the backslash again
					g_string_append_c (s,'\\');
			}
			g_string_append_c (s,c);
			quoted = FALSE;
		} else {
			if (c=='\\') quoted=TRUE;
			else
				g_string_append_c (s,c);
		}
	}

	return g_string_free (s, FALSE);
}

void
egg_markdown_clear(EggMarkdown *self)
{
	int i;

	for (i = 0; i < self->priv->link_table->len; i++) {
		g_free(g_array_index(self->priv->link_table, gchar *, i));
	}

	g_array_free(self->priv->link_table, TRUE);
	self->priv->link_table = g_array_new(FALSE, FALSE, sizeof(gchar *));

	for (i = 0; i < self->priv->code_table->len; i++) {
		g_free(g_array_index(self->priv->code_table, gchar *, i));
	}

	g_array_free(self->priv->code_table, TRUE);
	self->priv->code_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
}

/**
 * egg_markdown_to_text_line_format_sections:
	* @text: markdown text
.*
.* Format @text for the current output surface.
 * Call this function after egg_markdown_hold_protected.
 **/
static gchar *
egg_markdown_to_text_line_format_sections (EggMarkdown *self, const gchar *text)
{
	gchar *data = g_strdup (text);
	gchar *temp;

	/* smart quoting */
	if (self->priv->smart_quoting) {
		if (self->priv->escape) {
			temp = data;
			data = egg_markdown_to_text_line_formatter (temp, "&quot;", "“", "”");
			g_free (temp);

			temp = data;
			data = egg_markdown_to_text_line_formatter (temp, "&apos;", "‘", "’");
			g_free (temp);
		} else {
			temp = data;
			data = egg_markdown_to_text_line_formatter (temp, "\"", "“", "”");
			g_free (temp);

			temp = data;
			data = egg_markdown_to_text_line_formatter (temp, "'", "‘", "’");
			g_free (temp);
		}
	}


	/* bold-italic1 - this combo must be detected before the individual bold/italic */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "***", self->priv->tags.strong_em_start, self->priv->tags.strong_em_end);
	g_free (temp);

	/* bold-italic2 - this combo must be detected before the individual bold/italic */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "___", self->priv->tags.strong_em_start, self->priv->tags.strong_em_end);
	g_free (temp);

	/* bold1 */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "**", self->priv->tags.strong_start, self->priv->tags.strong_end);
	g_free (temp);

	/* bold2 */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "__", self->priv->tags.strong_start, self->priv->tags.strong_end);
	g_free (temp);

	/* italic1 */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "*", self->priv->tags.em_start, self->priv->tags.em_end);
	g_free (temp);

	/* italic2 */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "_", self->priv->tags.em_start, self->priv->tags.em_end);
	g_free (temp);

	/* strikethrough */
	temp = data;
	data = egg_markdown_to_text_line_formatter (temp, "---", self->priv->tags.strikethrough_start, self->priv->tags.strikethrough_end);
	g_free (temp);

	/* em-dash */
	temp = data;
	data = egg_markdown_replace (temp, " -- ", " — ", NULL);
	g_free (temp);

	/* remove redundant backslash */
	temp = data;
	data = egg_markdown_to_text_line_formatter_remove_backslash (temp);
	g_free (temp);

	return data;
}

/**
 * egg_markdown_add_pending:
 **/
static gboolean
egg_markdown_add_pending (EggMarkdown *self, const gchar *text)
{
	gchar *copy;

	if (self->priv->line_count >= self->priv->max_lines)
		return FALSE;

	/* https://spec.commonmark.org/0.30/#paragraphs */

	copy = g_strdup (text);

	if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK)
		g_strstrip (copy);

	g_string_append_printf (self->priv->pending, "%s%s", copy, PENDING_SEPARATOR);

	g_free (copy);
	return TRUE;
}

/**
 * egg_markdown_add_pending_br:
 **/
static gboolean
egg_markdown_add_pending_br (EggMarkdown *self, const gchar *line)
{
	GString *s;
	gboolean ret;

	s = g_string_new (line);
	g_strchomp (s->str); /* strip dangling spaces */
	g_string_set_size (s, strlen (s->str));

	/* erase pending separator and append hard-line-break keepsake */
	egg_markdown_string_chomp (s, PENDING_SEPARATOR, 1);
	g_string_insert_unichar (s, -1, iUNIPUA_BR);
	ret = egg_markdown_add_pending (self, s->str);

	g_string_free (s, TRUE);
	return ret;
}

/**
 * egg_markdown_add_pending_header:
 **/
static gboolean
egg_markdown_add_pending_header (EggMarkdown *self, const gchar *text)
{
	gchar *copy;
	gboolean ret;
	gchar *p;
	/*
	* https://spec.commonmark.org/0.30/#atx-heading
	*/

	/* right-strip #s iff preceded by space */
	p = copy = g_strdup (text);
	p += strlen(p) -1;
	while (p != copy && (*p == ' ' || *p == '\t')) /* optional trailing space */
		p--;
	while (p != copy && *p == '#')
		p--;
	if (*p == ' ' || *p == '\t')
		*p = '\0';

	ret = egg_markdown_add_pending (self, copy); /* strips spaces */
	g_free (copy);
	return ret;
}

/**
 * egg_markdown_count_chars_in_word:
 **/
static guint
egg_markdown_count_chars_in_word (const gchar *text, gchar find)
{
	guint i;
	guint len;
	guint count = 0;

	/* get length */
	len = strnlen (text, EGG_MARKDOWN_MAX_LINE_LENGTH);
	if (len == 0)
		goto out;

	/* find matching chars */
	for (i=0; i<len; i++) {
		if (text[i] == find)
			count++;
	}
out:
	return count;
}

/**
 * egg_markdown_code_word_boundary_start:
 * Find the word after the formatting prefix of input string.
 * Prefix is a possibly-empty sequence of formatting characters accepted
 * by egg_markdown_to_text_line_format_sections, e.g., '_', '*', '-'.
 * Prefix is the longest sequence of unescaped (with '\') formatting characters.
 * At most `len` characters of `s` are examined.
 * Return value: pointer to the string after prefix.
**/
static inline gchar *
egg_markdown_code_word_boundary_start (const gchar *str, size_t len)
{
	gboolean f;
	gint i;
	gchar *s = (gchar *) str;

	for (i = 0; i < len; i++, s++) {
		f = *s == '_' || *s == '*' || *s == '-';
		if (!f)
			break;
	}

	return s;
}

/**
 * egg_markdown_code_word_boundary_end:
 * Find the end of the word before the formatting suffix of input string.
 * Suffix is defined as in egg_markdown_code_word_boundary_start.
 * At most `len` characters of `s` are examined. Unicode PUA character
 * iUNIPUA_BR is treated as a word boundary.
 * Return value: pointer to 1-character string before suffix.
 * If suffix is empty the return value points to the last character of `s`.
**/
static inline gchar *
egg_markdown_code_word_boundary_end (const gchar *str, size_t len)
{
	gboolean f, b; /* previous char was backslash */
	gint i;
	gchar *e = NULL, *s = (gchar *) str;

	for (b = FALSE, i = 0; i < len; i++, s++) {
		f = *s == '_' || *s == '*' || *s == '-';
		if (f && !b) {
			if (e == NULL)
				e = s;
			continue;
		}
		/* '\357' = start of iUNIPUA_BR  */
		if (*s == '\357') {
			if (e == NULL)
				e = s;
			break;
		}
		e = NULL;
		if (*s == '\\')
			b = !b;
		else
			b = FALSE;
	}

	if (e)
		return e - 1;
	return (gchar *) str + (ptrdiff_t) (len - 1);
}

/**
 * egg_markdown_to_text_line_match_pair:
 * Match a pair of characters at the <s>tart and <e>nd of a string, ignoring
 * trailing punctuation only, e.g. "\"\"" to match quotes, "()" for brackets...
 * If the pair is matched, return the string length of string inside the pair.
 */
static inline gboolean
egg_markdown_to_text_line_match_pair (const gchar *s, gint *len, const gchar *pair)
{
	gboolean b = FALSE;
	gchar *e;

	if (*s == pair[0]) {
		for (e = (gchar *) s + *len - 1; e != s && *e != pair[1] && ispunct (*e); e--)
			;
		if (b = (e != s && *e == pair[1]))
			*len = (ptrdiff_t) (e - s) - 1;
	}
	return b;
}

/**
 * egg_markdown_word_is_ident:
 * Return: length of matched indentifier [[:alnum:]_]+
 * - first character not digit
 * - underscore can be escaped with backslash
 * @up: option uppercase only
**/
static inline gint
egg_markdown_word_is_ident (const gchar *s, size_t len, gboolean up)
{
	gint i;

	if (isdigit (*s))
		return 0;

	for (i = 0; i < len; i++) {
		switch (s[i]) {
			case '\\':
			case '_' :
				continue;
		}
		if (up ? isdigit (s[i]) || isupper (s[i]) : isalnum (s[i]))
			continue;
		break;
	}
	return i;
}

/**
 * egg_markdown_word_is_code:
 * Return value: TRUE if text can be code; for non-trivial cases,
 * also return dynamically allocated memory in *code, which holds
 * the code word surrounded by backticks; otherwise *code == NULL.
 * If text starts and/or ends with formatting characters
 * (cf. egg_markdown_to_text_line_format_sections), those will
 * prefix and/or suffix the code word that is returned in *code.
 * Allow backlash to escape a formatting character or itself.
 * Any unescaped backslash in return-word itself is removed.
**/
static gboolean
egg_markdown_word_is_code (const gchar *text, gchar **code)
{
	const gint len = strlen (text);
	gboolean b, f, ret = TRUE;       /* default return on goto out  */
	gchar *s, *e, *c = NULL;         /* return-word <s>tart and <e>nd */
	gint i, cwl;                     /* return-word length <s>...<e>  */

	/* find <s>tart of code word candidate after formatting prefix */
	const gchar *s0 = s = egg_markdown_code_word_boundary_start (text, len);
	/* find <e>nd of code word candidate before formatting suffix */
	const gchar *e0 = e = egg_markdown_code_word_boundary_end (text, len);

	cwl = (ptrdiff_t) (e - s) + 1;

/* Invariant: the return-word candidate <s>...<e> will be cwl chars long */
/* range of s [s0..e]; range of e [s..e0] */
/* Test Nth character before / after current return-word start / end */
#define LWING(N, C) (((ptrdiff_t)(s - s0) > 0) && (*(s - N) == (C)))
#define RWING(N, C) (((ptrdiff_t)(e0 - e) > 0) && (*(e + N) == (C)))

	/*************************************************************
	*                        code word tests                     *
	*************************************************************/

	*code = NULL;

	/* word is empty or too short or it is already part of other code */
	if (cwl <= 1) /* || *s == '`' || *e == '`') */
		goto not_code;

	/***********/
	/* dequote */
	/***********/
	if (egg_markdown_to_text_line_match_pair (s, &cwl, "\"\"") ||
		egg_markdown_to_text_line_match_pair (s, &cwl, "''")) {
			s++;
			e = s + cwl - 1;
	}

	/* URI can be non-ASCII */
	/* https://en.wikipedia.org/wiki/Internationalized_domain_name */
	{
		gchar **x, *ext[] = { "https://", "http://", "ftp://", NULL };
		guint len;
		for (x = ext; *x; x++) {
			len = strlen (*x);
			if (!strncmp (s, *x, len))
				goto out;

			/* [(<]http://...[>)] */
			if (cwl >= len + 2 && !strncmp (s + 1, *x, len - 1) &&
				(egg_markdown_to_text_line_match_pair (s, &cwl, "<>") ||
				egg_markdown_to_text_line_match_pair (s, &cwl, "()"))) {
					s++;
					e = s + cwl - 1;
					goto out;
				}
		}
	}

	/*******************************/
	/* ignore trailing punctuation */
	/*******************************/
	if (e[-1] != '\\') {
		for (i = cwl; i > 0 && ispunct (*e); )
			e--, i--;
		cwl = i;
	}

	/* absolute path len >= 4 bytes */
	if (cwl >= 4 && *s == '/') {
		if (RWING(1, '/'))
			e++, cwl++;
		goto out;
	}

	/******************************/
	/* ignore leading punctuation */
	/******************************/
	if (*s != '\\') {
		for (i = cwl; i > 0 && ispunct (*s); )
			s++, i--;
		cwl = i;
	}

	if (cwl <= 0) /* nothing to do */
		goto not_code;

	/* match file name != "" + extension */
	{
		gchar **x, *ext[] = { ".patch", ".diff", NULL };
		guint len;
		for (x = ext; *x; x++) {
			len = strlen (*x);
			if (cwl > len && !strncmp (e - len + 1, *x, len))
				goto out;
		}
	}

	/*************************************************************
	*                  ASCII only from here on                   *
	*************************************************************/

	for (i = 0; i < cwl; i++)
		if (!isascii(s[i]))
			goto not_code;

	/* bugzillas */
	if (LWING(1, '#')) {
		s--, cwl++;
		goto out;
	}

	/* function name followed by "()" no spaces */
	if (RWING(1, '(') && RWING(2, ')') &&
		egg_markdown_word_is_ident (s, cwl, FALSE) == cwl) {
		e = e + 2;
		cwl += 2;
		goto out;
	}

	/* email addresses */
	if (*s != '@' && *s != '.' && *e != '@') {
		guint dot = 0, at = 0;
		for (i = 0; i < cwl; i++)
			switch (s[i]) {
				case '.': ++dot; break;
				case '@': ++at;  break;
			}
		if (at == 1 && dot > 0)
			goto out;
	}

	/* uppercase identifier that includes at least one '_' */
	if (egg_markdown_word_is_ident (s, cwl, TRUE) == cwl && memchr (s, '_', cwl)) {
		if (LWING(1, '$'))
			s--, cwl++;
		goto out;
	}

not_code:
	ret = FALSE;
out:
	/* assemble code word for non-trivial cases */
	if (ret && *code == NULL && cwl > 0 && (cwl != len || strchr (s, '\\'))) {

		*code = g_malloc0 ((len + 3) * sizeof *c);

		/* formatting prefix */
		for (i = 0, c = (gchar *)text; c != s; c++)
			(*code)[i++] = *c;
		(*code)[i++] = '`';
		/* word - chop unescaped backslash */
		for (b = FALSE; c != e; c++)
			switch (*c) {
				case '\\':
					b = !b;
					if (b)
						continue;
				default:
					(*code)[i++] = *c;
					b = FALSE;
			}
		/* word end */
		(*code)[i++] = *e;
		(*code)[i++] = '`'; c++;
		/* formatting suffix */
		for (; *c; c++)
			(*code)[i++] = *c;
		(*code)[i] = '\0';
	}

	return ret;
}

/**
 * egg_markdown_word_auto_format_code_segment:
 * @text: markdown segment to be assessed for auto-codes
 * Return:
 * new auto-coded markdown segment or NULL if no auto-codes apply to @text
**/
static gchar *
egg_markdown_word_auto_format_code_segment (const gchar *text)
{

	guint i;
	gchar *temp = NULL;
	gchar **words;
	gchar *code;
	gboolean join = FALSE;

	/* split segment on spaces */
	words = g_strsplit (text, " ", -1);

	for (i = 0; words[i] != NULL; i++) {
		if (egg_markdown_word_is_code (words[i], &code)) {
			if (code)
				temp = code;
			else
				temp = g_strdup_printf ("`%s`", words[i]);
			g_free (words[i]);
			words[i] = temp;
			join = TRUE;
		}
	}
	if (join)
		temp = g_strjoinv (" ", words);
	g_strfreev (words);
	return temp;
}

/**
 * egg_markdown_hold_code_spans:
 * @text:  markdown text with protected segments, typically @self->pending
 *
 * Format @text for the current output surface.
 * Call this function between egg_markdown_hold_protected and release_line_endings.
 **/
static gchar *
egg_markdown_hold_code_spans (EggMarkdown *self, const gchar *text, const gboolean insert_autocodes)
{
	guint i;
	gchar *ret;
	GArray *slices = NULL;

#if PRINTERS > 1
	g_printerr (_SO"%s"_SE":\t\t\"%s\"\n", __FUNCTION__, text);
#endif
	/* identify inline code positions, and stash code text */
	slices = egg_markdown_slice_get_codes (self, text, TRUE);
	if (slices->len > 0) {
#if PRINTERS > 0
		_print_priv_table (self->priv->code_table,
			_SOyel"%s: CODE TABLE:"_SE" \"%s\"\n", __FUNCTION__, text);
#endif
		/* replace each CODE segment with code_ref to its stashed value */
		ret = egg_markdown_slice_join_codes (slices, text, insert_autocodes);
	} else {
		ret = g_strdup (text);
		if (insert_autocodes) {
			GString *buf = g_string_new ("");
#if PRINTERS > 0
			g_printerr (_SOyel"%s: AUTOCODE DISCOVERY:"_SE"\n", __FUNCTION__);
#endif
			_append_autocode (buf, ret, strlen (ret));
			g_free (ret);
			ret = g_string_free (buf, FALSE);
		}
	}
	g_array_free (slices, TRUE);

	return ret;
}

/**
 * egg_markdown_hold_protected:
 * @text: markdown text obtained from joining multiple input lines
 * Return:
 * newly-allocated string of markdown text in which some segments are
 * encoded to protect their content from mangling by character formatters.
 * For example, since the priority of markdown inline image is higher
 * than that of inline character formatting's, a markdown segment such as
 *    `![*image alt text*](image.png)`
 * must not be formatted as HTML
 *    `<src alt="<strong>alt text</strong>" src="image.png">`
 * also not as
 *    `<src alt="&lt;strong&gt;alt text&lt;/strong&gt;" src="image.png">`
 *
 * The return string encodes high priority segments as follows:
 * - includes PENDING_SEPARATOR and Unicode PUA characters.
 * - code_refs replace code spans (inline code segments of "`CODE`" markdown)
 * - code_refs replace discovered code spans (auto-code segments e.g. "/path/file")
 * - link_dest_refs replace link paths ("![](DEST)" and "[](DEST)" markdown)
 * - link_title_refs replace link titles (TITLE in '![](DEST "TITLE")' and '[](DEST "TITLE")')
 *
 * egg_markdown_to_text_line_formatter_code resolves code_refs
 * egg_markdown_to_text_line_formatter_link resolves link_dest_refs and link_title_refs.
 **/
static gchar *
egg_markdown_hold_protected (EggMarkdown *self, const gchar *text)
{
#ifdef PRINTERS
	g_printerr (_SOblu"%s: INPUT TEXT"_SE"\t\t\"%s\"\n", __FUNCTION__, text);
#endif
	guint i;
	gchar *temp = NULL, *atext = g_strdup (text);
	const gboolean insert_autocodes = egg_markdown_extension_enabled (self,
		EGG_MARKDOWN_EXTENSION_AUTOCODE) && self->priv->autocode;
	GArray *slices = NULL;

	/**********************************************************************
	*                             CODE SPANS                             *
	**********************************************************************/

	/* https://spec.commonmark.org/0.30/#example-341
	* Code span backticks have higher precedence than any other inline
	* constructs except HTML tags and autolinks. */

	atext = egg_markdown_hold_code_spans (self, temp = atext, insert_autocodes);
	g_free (temp);

	/* since autocode discovery could have inserted new CODE segments */
	/* hold the new CODE segments */
	if (self->priv->autocode) {
		atext = egg_markdown_hold_code_spans (self, temp = atext, FALSE);
		g_free (temp);
	}

	/**********************************************************************
	*                               LINKS                                *
	**********************************************************************/

	/* identify link positions, and stash link destinations */
	slices = egg_markdown_slice_get_links (self, atext, TRUE);
	if (slices->len > 0) {
		/* replace link destination with ref to its stashed value */
		atext = egg_markdown_slice_join_links (slices, temp = atext);
		g_free (temp);
#if PRINTERS > 0
		_print_priv_table (self->priv->link_table,
			_SOyel"%s: LINK TABLE:"_SE" \"%s\"\n", __FUNCTION__, atext);
#endif
	}
	g_array_free (slices, TRUE);

	return atext;
}

/**
 * _format_codeblock_line:
 * Helper for egg_markdown_flush_pending
 * @line is NULL at the end of the codeblock
 **/
static void
_format_codeblock_line (EggMarkdown *self, const gchar *line)
{
/* Commonmark 0.30 adds a line feed before the </code> closing tag, which
* creates an empty line at the end of the block. I don't like the empty
* line. I make it a compile-time option (default -U #undef).
*/
#ifndef COMMONMARK_CODEBLOCK_LINE_ENDING
#define FMT_CMCB1    "%s"
#define FMT_CMCB2    "\n%s"
#else
#define FMT_CMCB1    "\n%s"
#define FMT_CMCB2    "%s\n"
#endif
	static gboolean did_end = FALSE;
	const gchar *start, *end;
	gchar *copy = NULL;
	guint first_line_idx = self->priv->codeblock_start_line_idx;

	switch (self->priv->codeblock_mode) {
		case EGG_MARKDOWN_CODEBLOCK_TABLE:
			start = self->priv->tags.codeblock_table_start;
			end   = self->priv->tags.codeblock_table_end;
			break;
		case EGG_MARKDOWN_CODEBLOCK_SHEBANG:
			start = self->priv->tags.codeblock_shebang_start;
			end   = self->priv->tags.codeblock_shebang_end;
			break;
		case EGG_MARKDOWN_CODEBLOCK_STD:
			start = self->priv->tags.codeblock_start;
			end   = self->priv->tags.codeblock_end;
			break;
		case EGG_MARKDOWN_CODEBLOCK_FENCED:
			start = self->priv->tags.codeblock_start;
			end   = self->priv->tags.codeblock_end;
			++first_line_idx;
			break;
		default:
			/* shouldn't happen */
			g_printerr ("%s: internal error\n", __FUNCTION__);
			return;
	}

	if (self->priv->curr_line_idx == self->priv->codeblock_start_line_idx) {
		/* start block */
		g_string_append (self->priv->processed, start);
		did_end = FALSE;
	}
	if (line && !self->priv->eof) {
		/* first and subsequent lines */
		if (self->priv->output == EGG_MARKDOWN_OUTPUT_HTML)
			copy = g_markup_escape_text (line, -1); /* not needed for MarkdownTextView */
		else
			copy = (gchar *)line;

		if (self->priv->curr_line_idx == first_line_idx)
				/* first line only */
				g_string_append_printf (self->priv->processed, FMT_CMCB1, copy);
		else /* subsequent lines */
			if (self->priv->curr_line_idx > first_line_idx)
				g_string_append_printf (self->priv->processed, FMT_CMCB2, copy);
	}

	/* end properly closed block and run-away block */
	if (!did_end && (line == NULL || self->priv->eof)) {
		g_string_append_printf (self->priv->processed, "%s\n", end);
		did_end = TRUE;
	}

	if (copy != line)
		g_free (copy);
}

/**
 * egg_markdown_flush_pending:
 **/
static void
egg_markdown_flush_pending (EggMarkdown *self)
{
	gchar *copy = NULL;
	gchar *temp = NULL;

	/* no data yet */
	if (self->priv->mode == EGG_MARKDOWN_MODE_UNKNOWN)
		return;

	/* Right-trim at most 1 PENDING_SEPARATOR but not any white space before it */
	egg_markdown_string_chomp (self->priv->pending, PENDING_SEPARATOR, 1);

	/* nothing to flush */
	if (self->priv->pending->len == 0 && self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK)
		return;

	copy = g_strdup (self->priv->pending->str);

	/*
	*---------------------------------------------------------
	 * FIXME dirty pot extraction and gettext lookup due to
	 * embedded code_refs, link_dest_refs and link_title_refs.
	*---------------------------------------------------------
	*/
	/* check for translation first */
	if (self->priv->pot_output || self->priv->use_gettext) {
		copy = egg_markdown_to_text_line_formatter_gettext (self, temp = copy);
		g_free(temp);
	}

	/* then check for exec, before anything, and so its output can be interpreted by others */
	if (self->priv->exec) {
		copy = egg_markdown_to_text_line_formatter_exec (self, temp = copy);
		g_free(temp);
	}

	/* pango requires escaping */
	if (!self->priv->escape && self->priv->output == EGG_MARKDOWN_OUTPUT_PANGO) {
		g_strdelimit (copy, "<", '(');
		g_strdelimit (copy, ">", ')');
	}

	/* protect code, link, and image segments */
	if (self->priv->mode == EGG_MARKDOWN_MODE_PARA ||
	     self->priv->mode == EGG_MARKDOWN_MODE_BULLETT) {
		copy = egg_markdown_hold_protected (self, temp = copy);
		g_free (temp);
	}

	/* escape */
	if (self->priv->escape) {
		copy = g_markup_escape_text (temp = copy, -1);
		g_free (temp);
	}

	/*
	* format inline elements
	*/

	if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK) {
		guint count = 0;

		/* last chance to match multi-line code spans, per commonmark spec */
		/* since line endings are still protected as iUNIPUA_BR.       */
			copy = egg_markdown_hold_code_spans (self, temp = copy, FALSE);
			g_free (temp);

	release_line_endings:
		/* replace protected line endings with output-surface-specific endings */
		copy = egg_markdown_replace_unichars (temp = copy, iUNIPUA_BR, PENDING_SEPARATOR, self->priv->tags.br, &count);
		g_free (temp);
		self->priv->line_count += count; /* output line count - not used */

		copy = egg_markdown_to_text_line_format_sections (self, temp = copy);
		g_free (temp);

#ifdef PRINTERS
		g_printerr (_SO"%s releasing:"_SE"\t\"%s\"\n", __FUNCTION__, copy);
#endif
	/* format inline protected code spans */
	release_code_spans:
		copy = egg_markdown_to_text_line_formatter_code (self, temp = copy);
		g_free (temp);

	/* replace line endings that emerged from code spans with spaces */
	/* per commonmark spec */
	release_emerged_line_endings:
		copy = egg_markdown_replace_unichars (temp = copy, iUNIPUA_BR, NULL, "  ", NULL);
		g_free (temp);

	/* format inline protected links (including images) */
	release_links_and_images:
		copy = egg_markdown_to_text_line_formatter_link (self, temp = copy);
		g_free(temp);
	}

	temp = NULL;
	DEBUG ("adding '%s'", copy);

	/*
	* format block elements
	*/

	if (self->priv->mode == EGG_MARKDOWN_MODE_BULLETT) {
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.bullett_start, copy, self->priv->tags.bullett_end);
		self->priv->line_count++;
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_H1) {
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h1_start, copy, self->priv->tags.h1_end);
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_H2) {
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h2_start, copy, self->priv->tags.h2_end);
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_H3) {
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.h3_start, copy, self->priv->tags.h3_end);
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK) {
		_format_codeblock_line (self, copy);
		self->priv->line_count++;
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_PARA) {
		g_string_append_printf (self->priv->processed, "%s%s%s\n", self->priv->tags.para_start, copy, self->priv->tags.para_end);
		self->priv->line_count++;
	} else if (self->priv->mode == EGG_MARKDOWN_MODE_RULE) {
		g_string_append_printf (self->priv->processed, "%s\n", copy);
		self->priv->line_count++;
	}
	g_free (copy);

	/* clear */
	g_string_truncate (self->priv->pending, 0);
}

/**
 * egg_markdown_enter_codeblock_mode:
**/
static inline void
egg_markdown_enter_codeblock_mode (EggMarkdown *self, EggMarkdownCodeblockMode codeblock_mode)
{
	self->priv->mode = EGG_MARKDOWN_MODE_CODEBLOCK;
	self->priv->codeblock_mode = codeblock_mode;
	self->priv->codeblock_prev_exec = self->priv->exec;
	self->priv->exec = FALSE;
	if (self->priv->codeblock_syntax != EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT) {
		self->priv->codeblock_prev_pot = self->priv->temporary_disable_pot;
		self->priv->temporary_disable_pot = TRUE;
	}
	self->priv->codeblock_start_line_idx = self->priv->curr_line_idx;
}

/**
 * egg_markdown_exit_codeblock_mode:
 * Return: `TRUE`, which means that we have set `self->priv->mode`.
**/
static inline gboolean
egg_markdown_exit_codeblock_mode (EggMarkdown *self)
{
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	self->priv->exec = self->priv->codeblock_prev_exec;
	if (self->priv->codeblock_syntax != EGG_MARKDOWN_CODEBLOCK_SYNTAX_GETTEXT)
		self->priv->temporary_disable_pot = self->priv->codeblock_prev_pot;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->codeblock_start_line_idx = -1;
	return TRUE;
}

/**
 * egg_markdown_is_codeblock_mode:
 **/
static inline gboolean
egg_markdown_is_codeblock_mode (EggMarkdown *self, EggMarkdownCodeblockMode codeblock_mode)
{
	return self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK &&
		self->priv->codeblock_mode == codeblock_mode;
}

/**
 * egg_markdown_to_text_line_process:
 * @self: the EggMarkdown instance
 * @line: current markdown line
 * @lines: NULL-terminated array of all markdown lines
 * @linenum: current line index in @lines[]
 **/
static gboolean
egg_markdown_to_text_line_process (EggMarkdown *self, const gchar *line, const gchar *next_line)
{
	gboolean ret = FALSE;

	/* HTML comment block */
	/* if we are already in comment block mode, skip everything and only check for end-of-comment */
	if (self->priv->mode == EGG_MARKDOWN_MODE_COMMENTBLOCK) {
		if (egg_markdown_to_text_line_is_blockcomment_end(line)) {
			self->priv->mode = self->priv->prev_mode;
		}
		ret = TRUE;
		goto out;
	}

	/* extension: shebang code block (whole file) */
	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_SHEBANG)) {
		if (self->priv->autocode_shebang || // reset on (re-)parsing rendered file
			( (self->priv->curr_line_idx == 0) && // only look for shebang on first line
			  egg_markdown_to_text_line_is_shebang (line) )
			) {
			if (!self->priv->autocode_shebang)
				egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_SHEBANG);
			ret = self->priv->autocode_shebang = TRUE;
			self->priv->mode = EGG_MARKDOWN_MODE_CODEBLOCK;
			self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_SHEBANG;

			/* render codeblock */
			egg_markdown_add_pending (self, line);
			egg_markdown_flush_pending (self); // end the line
			goto out;
		}
	}

	/* extension: pipe table as code block - one-line lookahead */
	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_TABLE)) {
		if (egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE) &&
			egg_markdown_to_text_line_is_table_end (next_line, self->priv->table_columns))
		{
		/* end table code block */
			egg_markdown_add_pending (self, line);
			egg_markdown_flush_pending (self);
			_format_codeblock_line (self, NULL);
			ret = egg_markdown_exit_codeblock_mode (self);
			goto out;
		}
		if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK && !egg_markdown_is_codeblock_mode(self, EGG_MARKDOWN_CODEBLOCK_TABLE)) {
			if (ret = egg_markdown_to_text_line_is_pipe_table_start (
				line, next_line, &self->priv->table_columns))
			{
		/* begin table block */
				egg_markdown_flush_pending (self);
				egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE);
				/* don't _format_codeblock_line (self, line); */
			}
		}
		/* flush each line while EGG_MARKDOWN_CODEBLOCK_TABLE */
		if (ret = egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_TABLE))
		{
			egg_markdown_add_pending (self, line);
			egg_markdown_flush_pending (self);
			goto out;
		}
	}

	/* extension: fenced code block */
	if (egg_markdown_extension_enabled (self, EGG_MARKDOWN_EXTENSION_GITHUB)) {
		if (self->priv->codeblock_mode != EGG_MARKDOWN_CODEBLOCK_TABLE) {
			if (ret = egg_markdown_to_text_line_is_fenced_code (line, &self->priv->codeblock_syntax)) {
				if (self->priv->codeblock_mode != EGG_MARKDOWN_CODEBLOCK_FENCED) {
					// begin code block
					egg_markdown_flush_pending (self);
					egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_FENCED);
					_format_codeblock_line (self, line);
				} else {
					// end code block
					_format_codeblock_line (self, NULL);
					ret = egg_markdown_exit_codeblock_mode (self);
				}
				goto out;
			}
			/* flush each line while EGG_MARKDOWN_CODEBLOCK_FENCED */
			if (ret = egg_markdown_is_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_FENCED))
			{
				egg_markdown_add_pending (self, line);
				egg_markdown_flush_pending (self);
				goto out;
			}
		}
	}

	/* indented code block */
	if (self->priv->mode != EGG_MARKDOWN_MODE_PARA &&
		(ret = egg_markdown_to_text_line_is_indented_code (self, line))) {
		const gchar *start = line + 4;

		if (self->priv->mode != EGG_MARKDOWN_MODE_CODEBLOCK) {
			// begin indented code block
			egg_markdown_flush_pending (self);
			egg_markdown_enter_codeblock_mode (self, EGG_MARKDOWN_CODEBLOCK_STD);
			/* don't _format_codeblock_line (self, line); */
		}
		/* flush each line while EGG_MARKDOWN_CODEBLOCK_STD */
		if (*line == '\t') {
			start = line + 1;
			egg_markdown_add_pending (self, "    ");
		}
		egg_markdown_add_pending (self, start);
		egg_markdown_flush_pending (self);
		goto out;
	}
	else if (self->priv->mode == EGG_MARKDOWN_MODE_CODEBLOCK) {
		/* end indented code block */
		_format_codeblock_line (self, NULL);
		ret = egg_markdown_exit_codeblock_mode (self);
	}

	/* HTML comment block or one-line comment span */
	/* when we get here we can be sure we are not inside a comment block */
	if (ret = egg_markdown_to_text_line_is_blockcomment_start(line)) {
		if (egg_markdown_to_text_line_is_blockcomment_end(line)) {
			/* whole line is a comment: throw away the line without setting any mode */
		} else {
			/* stash mode - will be restored when comment block ends */
			self->priv->prev_mode = self->priv->mode;
			self->priv->mode = EGG_MARKDOWN_MODE_COMMENTBLOCK;
		}
		goto out;
	}

	/* <a name=...> - throw away the line */
	if (ret = egg_markdown_to_text_line_is_tag_a_name(line)) {
		/* throw away the line unless output mode is HTML; stay in current mode */
		if (self->priv->output != EGG_MARKDOWN_OUTPUT_HTML)
			goto out;
	}

	/* blank */
	ret = egg_markdown_to_text_line_is_blank (line);
	if (ret) {
		DEBUG ("blank: '%s'", line);
		egg_markdown_flush_pending (self);
		/* a new line after a list is the end of list, not a gap */
		if (self->priv->mode != EGG_MARKDOWN_MODE_BULLETT)
			ret = egg_markdown_add_pending (self, "\n");
		self->priv->mode = EGG_MARKDOWN_MODE_BLANK;
		goto out;
	}

	/* header1_type2 */
	ret = egg_markdown_to_text_line_is_header1_type2 (line);
	if (ret && self->priv->mode == EGG_MARKDOWN_MODE_PARA) {
		DEBUG ("header1_type2: '%s'", line);
		self->priv->mode = EGG_MARKDOWN_MODE_H1;
		goto out;
	}

	/* header2_type2 */
	ret = egg_markdown_to_text_line_is_header2_type2 (line);
	if (ret && self->priv->mode == EGG_MARKDOWN_MODE_PARA) {
		DEBUG ("header2_type2: '%s'", line);
		self->priv->mode = EGG_MARKDOWN_MODE_H2;
		goto out;
	}

	/* rule */
	ret = egg_markdown_to_text_line_is_rule (line);
	if (ret) {
		DEBUG ("rule: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_RULE;
		ret = egg_markdown_add_pending (self, self->priv->tags.rule);
		goto out;
	}

	/* bullett */
	ret = egg_markdown_to_text_line_is_bullett (line);
	if (ret) {
		DEBUG ("bullett: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_BULLETT;
		if (egg_markdown_to_text_line_has_line_break (line))
			ret = egg_markdown_add_pending_br (self, &line[2]);
		else
			ret = egg_markdown_add_pending (self, &line[2]);
		goto out;
	}

	/* header1 */
	ret = egg_markdown_to_text_line_is_header1 (line);
	if (ret) {
		DEBUG ("header1: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_H1;
		if (egg_markdown_to_text_line_has_line_break (line))
			ret = egg_markdown_add_pending_br (self, &line[2]);
		else
			ret = egg_markdown_add_pending_header (self, &line[2]);
		goto out;
	}

	/* header2 */
	ret = egg_markdown_to_text_line_is_header2 (line);
	if (ret) {
		DEBUG ("header2: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_H2;
		if (egg_markdown_to_text_line_has_line_break (line))
			ret = egg_markdown_add_pending_br (self, &line[3]);
		else
			ret = egg_markdown_add_pending_header (self, &line[3]);
		goto out;
	}

	/* header3 */
	ret = egg_markdown_to_text_line_is_header3 (line);
	if (ret) {
		DEBUG ("header3: '%s'", line);
		egg_markdown_flush_pending (self);
		self->priv->mode = EGG_MARKDOWN_MODE_H3;
		if (egg_markdown_to_text_line_has_line_break (line))
			ret = egg_markdown_add_pending_br (self, &line[4]);
		else
			ret = egg_markdown_add_pending_header (self, &line[4]);
		goto out;
	}

	/* directives */
	ret = egg_markdown_to_text_line_is_textdomain (line);
	if (ret) {
		egg_markdown_flush_pending (self); // end line
		gchar *domain = g_strdup(&line[sizeof(TEXTDOMAIN_DIRECTIVE)-1]);
		g_strstrip (domain);
		if (domain[0]) {
			g_debug ("text_domain: '%s'\n", domain);
			gchar *dir = bindtextdomain(textdomain(NULL),NULL);
			textdomain(domain);
			bindtextdomain(textdomain(NULL), dir);
		}
		g_free (domain);
		goto out;
	}

	ret = egg_markdown_to_text_line_is_exec (line);
	if (ret) {
		egg_markdown_flush_pending (self); // end line
		gchar *tmp = g_strdup(&line[sizeof(EXEC_DIRECTIVE)-1]);
		g_strstrip (tmp);
		if (tmp[0]) {
			g_debug ("exec: '%s'\n", tmp);
			if (g_str_has_prefix(tmp, "yes") && self->priv->global_exec)
				self->priv->exec = TRUE;
			else
				self->priv->exec = FALSE;
		}
		g_free (tmp);
		goto out;
	}

	ret = egg_markdown_to_text_line_is_nopot (line);
	if (ret) {
		egg_markdown_flush_pending (self); // end line
		gchar *tmp = g_strdup(&line[sizeof(NOPOT_DIRECTIVE)-1]);
		g_strstrip (tmp);
		if (tmp[0]) {
			g_debug ("nopot: '%s'\n", tmp);
			if (g_str_has_prefix(tmp, "yes"))
				self->priv->temporary_disable_pot = TRUE;
			else
				self->priv->temporary_disable_pot = FALSE;
		}
		g_free (tmp);
		goto out;
	}


para:
	/* paragraph */
	/* hard line-break */
	ret = egg_markdown_to_text_line_has_line_break (line);
	if (ret || self->priv->mode == EGG_MARKDOWN_MODE_BLANK || self->priv->mode == EGG_MARKDOWN_MODE_UNKNOWN) {
		if (self->priv->mode != EGG_MARKDOWN_MODE_PARA) {
			/* start new paragraph */
			egg_markdown_flush_pending (self);
			self->priv->mode = EGG_MARKDOWN_MODE_PARA;
		}
		if (ret) {
			ret = egg_markdown_add_pending_br (self, line);
			goto out;
		}
	}

	/* add to pending */
	DEBUG ("continue: '%s'", line);
	ret = egg_markdown_add_pending (self, line);
out:
	/* if we failed to add, we don't know the mode */
	if (!ret)
		self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	return ret;
}

/**
 * egg_markdown_linkbuilder_pango:
 * @text: can be NULL
 * @dest: can be NULL, ignored
 * @title: can be NULL, ignored
 * @link_dest_id: can be -1;
 **/
static gchar *
egg_markdown_linkbuilder_pango (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	/*
	* Here and in egg_markdown_imagebuilder_pango(), We need to pass some
	* extra data directly to the output surface renderer, in this case the
	* code that uses GtkTextView in markdown-text-view.c.  Since extending
	* Pango markup to allow new tags is too complicated, our extra data will
	* piggyback the Pango markup "font" attribute. The renderer will parse
	* the font string and call egg_markdown_get_link_dest() et al. to
	* obtain the link DEST.
	*
	* link_table index embedded in pango font description attribute:
	*   font="@dest=<type><id>",   with
	*   <type> "Lu"(URI), "Ip"(image path); <id> int(link table index aka id)
	*
	* Is it safe to piggyback the font description attribute{1}?
	* Test <span>s with pango-view --markup -t, e.g.
	*   font="Serif Bold Italic 44 @x=1,y=3,z=4"
	* and play with "Serif", "Bold", "Italic", "44"...
	*
	* @..., if any, is called VARIATIONS and must start with "@" in tail position.
	* I tested made-up variations; they are allowed and preserve the font.
	*
	* {1} https://docs.gtk.org/Pango/type_func.FontDescription.from_string.html#description
	* {2} https://gitlab.gnome.org/GNOME/pango/-/blob/main/pango/fonts.c#L1252
	*     is the "[src]" link in {1}
	*/
	gchar *ret   = NULL;
	gchar *style = "foreground=\"blue\" underline=\"single\"";

	if (dest && link_dest_id >= 0)
		ret = g_strdup_printf (
			"<span font=\"@%s%d\" %s>%s</span>%s%s%s",
			_tag_info[EGG_MARKDOWN_TAG_DEST_LINK_URI_ID],
			link_dest_id, style, text ? text : "⯅⯅",
			title ? " (" : "", title ? title : "", title ? ")" : "");

	else /* valid commonmark markdown w/o dest, e.g. [text]() - I know... */
		ret = g_strdup_printf (
			"<span %s>%s</span>%s%s%s", style, text ? text : "↑",
			title ? " (" : "", title ? title : "", title ? ")" : "");
	return ret;
}

/**
 * egg_markdown_linkbuilder_html
 * @text: can be NULL
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_linkbuilder_html (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *ret, *esc_title = NULL;

	if (title)
		esc_title = g_markup_escape_text (title, -1);
	if (esc_title)
		ret = g_strdup_printf("<a href=\"%s\" title=\"%s\">%s</a>", dest ? dest : "", esc_title, text ? text : "");
	else
		ret = g_strdup_printf("<a href=\"%s\">%s</a>", dest ? dest : "", text ? text : "");
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_linkbuilder_text
 * @text: can be NULL
 * @dest: can be NULL
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_linkbuilder_text (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	if (title)
		return g_strdup_printf ("%s%s%s%s%s%s(%s)", text ? text : "", text ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "",
				text || dest ? " " : "", title);
	else
		return g_strdup_printf ("%s%s%s%s%s", text ? text : "", text ? " " : "",
				dest ? "<" : "", dest ? dest : "", dest ? ">" : "");
}

/**
 * egg_markdown_imagebuilder_pango:
 * @text: can be NULL, image alternate text
 * @dest: can be NULL, image path, ignored
 * @title: can be NULL, ignored
 * @link_dest_id: can be -1
 **/
static gchar *
egg_markdown_imagebuilder_pango (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *style = "underline=\"double\"";
	char alt[16] = {0};
	/*
	* Text in a GtkTextBuffer can be marked with tags. A tag is an attribute that
	* can be applied to some range of text. Here we create the tag with pango markup.
	* GtkTextView tags don't exist by themselves: they exist together with text != "".
	*/

	/* Arbitrarily, make empty text fall back to the stringified link_dest_id */
	if (text == NULL)
		snprintf (alt, 16, "%d", link_dest_id);

	/* See comment in egg_markdown_linkbuilder_pango() */
	return g_strdup_printf ("<span font=\"@%s%d\" %s>%s</span>",
		_tag_info[EGG_MARKDOWN_TAG_DEST_IMAGE_PATH_ID],
		link_dest_id, style, text ? text : alt);
}

/**
 * egg_markdown_imagebuilder_html
 * @text: can be NULL, image alternate text
 * @dest: can be NULL, image path
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_imagebuilder_html (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	gchar *ret, *esc_text = NULL, *esc_title = NULL;
	if (text)
		esc_text = g_markup_escape_text (text, -1);
	if (title)
		esc_title = g_markup_escape_text (title, -1);
	if (esc_title)
		ret = g_strdup_printf ("<img src=\"%s\" alt=\"%s\" title=\"%s\" />", dest ? dest : "", esc_text ? esc_text : "", esc_title);
	else
		ret = g_strdup_printf ("<img src=\"%s\" alt=\"%s\" />", dest ? dest : "", esc_text ? esc_text : "");
	g_free (esc_text);
	g_free (esc_title);
	return ret;
}

/**
 * egg_markdown_imagebuilder_text
 * @text: can be NULL, image alternate text
 * @dest: can be NULL, image path, ignored
 * @title: can be NULL
 * @link_dest_id: ignored
 **/
static gchar *
egg_markdown_imagebuilder_text (const gchar *text, const gchar *dest, const gchar *title, const gint link_dest_id)
{
	if (title)
		return g_strdup_printf ("%s%s(%s)", text ? text : "", text ? " " : "", title);
	else
		return g_strdup (text ? text : "");
}

/**
 * egg_markdown_set_output:
 **/
gboolean
egg_markdown_set_output (EggMarkdown *self, EggMarkdownOutput output)
{
	gboolean ret = TRUE;
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);

	/* PangoMarkup */
	if (output == EGG_MARKDOWN_OUTPUT_PANGO) {
		self->priv->tags.em_start = "<i>";
		self->priv->tags.em_end = "</i>";
		self->priv->tags.strong_start = "<b>";
		self->priv->tags.strong_end = "</b>";
		self->priv->tags.strong_em_start = "<b><i>";
		self->priv->tags.strong_em_end = "</i></b>";
		self->priv->tags.code_start = "<tt><span bgcolor=\"#eee\">";
		self->priv->tags.code_end = "</span></tt>";
		self->priv->tags.codeblock_start = "<tt><span bgcolor=\"#eee\">";
		self->priv->tags.codeblock_end = "</span></tt>";
		self->priv->tags.codeblock_table_start = "<tt><span>";
		self->priv->tags.codeblock_table_end = "</span></tt>";
		self->priv->tags.codeblock_shebang_start = "<tt><span>";
		self->priv->tags.codeblock_shebang_end = "</span></tt>";
		self->priv->tags.strikethrough_start = "<s>";
		self->priv->tags.strikethrough_end = "</s>";
		self->priv->tags.h1_start = "<span color=\"#444\" size=\"xx-large\"><b>";
		self->priv->tags.h1_end = "</b></span>";
		self->priv->tags.h2_start = "<big><big><b>";
		self->priv->tags.h2_end = "</b></big></big>";
		self->priv->tags.h3_start = "<big><span variant=\"smallcaps\"><b>";
		self->priv->tags.h3_end = "</b></span></big>";
		self->priv->tags.bullett_start = "  • ";
		self->priv->tags.bullett_end = "";
		self->priv->tags.rule = "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_pango;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_pango;

	/* XHTML */
	} else if (output == EGG_MARKDOWN_OUTPUT_HTML) {
		self->priv->tags.em_start = "<em>";
		self->priv->tags.em_end = "</em>";
		self->priv->tags.strong_start = "<strong>";
		self->priv->tags.strong_end = "</strong>";
		self->priv->tags.strong_em_start = "<strong><em>";
		self->priv->tags.strong_em_end = "</em></strong>";
		self->priv->tags.code_start = "<code>";
		self->priv->tags.code_end = "</code>";
		self->priv->tags.codeblock_start = "<pre><code>";
		self->priv->tags.codeblock_end = "</code></pre>";
		self->priv->tags.codeblock_table_start = "<pre><code>";
		self->priv->tags.codeblock_table_end = "</code></pre>";
		self->priv->tags.codeblock_shebang_start = "<pre><code>";
		self->priv->tags.codeblock_shebang_end = "</code></pre>";
		self->priv->tags.strikethrough_start = "<strike>";
		self->priv->tags.strikethrough_end = "</strike>";
		self->priv->tags.h1_start = "<h1>";
		self->priv->tags.h1_end = "</h1>";
		self->priv->tags.h2_start = "<h2>";
		self->priv->tags.h2_end = "</h2>";
		self->priv->tags.h3_start = "<h3>";
		self->priv->tags.h3_end = "</h3>";
		self->priv->tags.bullett_start = "<li>";
		self->priv->tags.bullett_end = "</li>";
		self->priv->tags.rule = "<hr>";
		self->priv->tags.para_start = "<p>";
		self->priv->tags.para_end = "</p>";
		self->priv->tags.br = "<br/>\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_html;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_html;

	/* tty - use Linux/vt100 codes */
	} else if (output == EGG_MARKDOWN_OUTPUT_TTY) {
#define ESCP "\033"
		self->priv->tags.em_start = ESCP "[4m";
		self->priv->tags.em_end = ESCP "[24m";
		self->priv->tags.strong_start = ESCP "[1m";
		self->priv->tags.strong_end = ESCP "[21m";
		self->priv->tags.strong_em_start = ESCP "[1m" ESCP "[4m";
		self->priv->tags.strong_em_end = ESCP "[24m" ESCP "[21m";
		self->priv->tags.code_start = ESCP "[36m";
		self->priv->tags.code_end = ESCP "[39m";
		self->priv->tags.codeblock_start = ESCP "[36m";
		self->priv->tags.codeblock_end = ESCP "[39m";
		self->priv->tags.codeblock_table_start = "";
		self->priv->tags.codeblock_table_end = "";
		self->priv->tags.codeblock_shebang_start = "";
		self->priv->tags.codeblock_shebang_end = "";
		self->priv->tags.strikethrough_start = "--";
		self->priv->tags.strikethrough_end = "--";
		self->priv->tags.h1_start = ESCP "[7;1m";
		self->priv->tags.h1_end = ESCP "[27;21m";
		self->priv->tags.h2_start = ESCP "[7m";
		self->priv->tags.h2_end = ESCP "[27m";
		self->priv->tags.h3_start = ESCP "[1m"; // same as bold
		self->priv->tags.h3_end = ESCP "[21m";
		self->priv->tags.bullett_start = "* ";
		self->priv->tags.bullett_end = "";
		self->priv->tags.rule = " ----- \n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_text;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_text;

	/* plain text */
	} else if (output == EGG_MARKDOWN_OUTPUT_TEXT) {
		self->priv->tags.em_start = "";
		self->priv->tags.em_end = "";
		self->priv->tags.strong_start = "";
		self->priv->tags.strong_end = "";
		self->priv->tags.code_start = "";
		self->priv->tags.code_end = "";
		self->priv->tags.codeblock_start = "";
		self->priv->tags.codeblock_end = "";
		self->priv->tags.codeblock_table_start = "";
		self->priv->tags.codeblock_table_end = "";
		self->priv->tags.codeblock_shebang_start = "";
		self->priv->tags.codeblock_shebang_end = "";
		self->priv->tags.strikethrough_start = "";
		self->priv->tags.strikethrough_end = "";
		self->priv->tags.h1_start = "[";
		self->priv->tags.h1_end = "]";
		self->priv->tags.h2_start = "-";
		self->priv->tags.h2_end = "-";
		self->priv->tags.h3_start = "~";
		self->priv->tags.h3_end = "~";
		self->priv->tags.bullett_start = "* ";
		self->priv->tags.bullett_end = "";
		self->priv->tags.rule = " ----- \n";
		self->priv->tags.para_start = "";
		self->priv->tags.para_end = "";
		self->priv->tags.br = "\n";
		self->priv->tags.link_builder = egg_markdown_linkbuilder_text;
		self->priv->tags.image_builder = egg_markdown_imagebuilder_text;

	/* unknown */
	} else {
		g_warning ("unknown output enum");
		ret = FALSE;
	}

	/* save if valid */
	if (ret)
		self->priv->output = output;
	return ret;
}

/**
 * egg_markdown_set_max_lines:
 **/
gboolean
egg_markdown_set_max_lines (EggMarkdown *self, gint max_lines)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->max_lines = max_lines;
	return TRUE;
}

/**
 * egg_markdown_set_smart_quoting:
 **/
gboolean
egg_markdown_set_smart_quoting (EggMarkdown *self, gboolean smart_quoting)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->smart_quoting = smart_quoting;
	return TRUE;
}

/**
 * egg_markdown_set_escape:
 **/
gboolean
egg_markdown_set_escape (EggMarkdown *self, gboolean escape)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->escape = escape;
	return TRUE;
}

/**
 * egg_markdown_set_autocode:
 **/
gboolean
egg_markdown_set_autocode (EggMarkdown *self, gboolean autocode)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->autocode = autocode;
	return TRUE;
}

/**
 * egg_markdown_set_exec:
 **/
gboolean
egg_markdown_set_exec (EggMarkdown *self, gboolean exec)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->exec = self->priv->global_exec = exec;
	return TRUE;
}

/**
 * egg_markdown_set_pot_output:
 **/
gboolean
egg_markdown_set_pot_output (EggMarkdown *self, FILE *f, gchar *input_filename)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->pot_output = f;
	self->priv->input_filename = g_strdup(input_filename);
	return TRUE;
}

/**
 * egg_markdown_set_use_gettext:
 **/
gboolean
egg_markdown_set_use_gettext (EggMarkdown *self, gboolean use_gettext)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->use_gettext = use_gettext;
	return TRUE;
}

/**
 * egg_markdown_set_extensions:
 **/
gboolean
egg_markdown_set_extensions (EggMarkdown *self, const EggMarkdownExtensions ext)
{
	g_return_val_if_fail (EGG_IS_MARKDOWN (self), FALSE);
	self->priv->extensions = ext;
	return TRUE;
}

/**
 * egg_markdown_parse:
 **/
gchar *
egg_markdown_parse (EggMarkdown *self, const gchar *markdown)
{
	gchar **lines;
	guint i;
	guint last_line_idx;
	gchar *temp;
	gboolean ret;

	g_return_val_if_fail (EGG_IS_MARKDOWN (self), NULL);
	g_return_val_if_fail (self->priv->output != EGG_MARKDOWN_OUTPUT_UNKNOWN, NULL);

	DEBUG ("input='%s'", markdown);

	/* process */
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->codeblock_start_line_idx = -1;
	self->priv->line_count = 0;
	self->priv->curr_line_idx = -1;
	self->priv->eof = FALSE;
	g_string_truncate (self->priv->pending, 0);
	g_string_truncate (self->priv->processed, 0);
	lines = g_strsplit (markdown, "\n", -1);
	last_line_idx = g_strv_length (lines) - 1;
	self->priv->autocode_shebang = FALSE;

	/* process each line */
	for (i = 0; (self->priv->curr_line_idx = i) <= last_line_idx; ++i) {
		ret = egg_markdown_to_text_line_process (self, lines[i], i < last_line_idx ? lines[i + 1] : NULL);
		if (!ret)
			break;
	}
	g_strfreev (lines);
	self->priv->eof = TRUE;
	egg_markdown_flush_pending (self);

	/* remove trailing \n */
	egg_markdown_string_chomp (self->priv->processed, "\n", 1);

	/* get a copy */
	temp = g_strdup (self->priv->processed->str);
	g_string_truncate (self->priv->pending, 0);
	g_string_truncate (self->priv->processed, 0);

	DEBUG ("output='%s'", temp);

	return temp;
}

/**
 * egg_markdown_class_init:
 * @klass: The EggMarkdownClass
 **/
static void
egg_markdown_class_init (EggMarkdownClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	object_class->finalize = egg_markdown_finalize;
G_GNUC_BEGIN_IGNORE_DEPRECATIONS;
	g_type_class_add_private (klass, sizeof (EggMarkdownPrivate));
G_GNUC_END_IGNORE_DEPRECATIONS;
}

/**
 * egg_markdown_init:
 **/
static void
egg_markdown_init (EggMarkdown *self)
{
	self->priv = EGG_MARKDOWN_GET_PRIVATE (self);

	self->priv->extensions = 0xffff; // enable all extensions
	self->priv->mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->prev_mode = EGG_MARKDOWN_MODE_UNKNOWN;
	self->priv->codeblock_mode = EGG_MARKDOWN_CODEBLOCK_NONE;
	self->priv->codeblock_syntax = EGG_MARKDOWN_CODEBLOCK_SYNTAX_NONE;
	self->priv->output = EGG_MARKDOWN_OUTPUT_UNKNOWN;
	self->priv->pending = g_string_new ("");
	self->priv->processed = g_string_new ("");
	self->priv->regex_table = g_array_set_size(g_array_sized_new(FALSE, TRUE, sizeof (GRegex *), EGG_MARKDOWN_REGEX_LEN), EGG_MARKDOWN_REGEX_LEN);
	self->priv->code_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
	self->priv->link_table = g_array_new(FALSE, FALSE, sizeof(gchar *));
	self->priv->max_lines = -1;
	self->priv->smart_quoting = FALSE;
	self->priv->escape = FALSE;
	self->priv->autocode = FALSE;
	self->priv->autocode_shebang = FALSE;
	self->priv->exec = self->priv->global_exec = FALSE;
	self->priv->pot_output = NULL;
	self->priv->temporary_disable_pot = FALSE;
	self->priv->use_gettext = FALSE;
	self->priv->pot_output = NULL;
	self->priv->input_filename = NULL;
}

/**
 * egg_markdown_finalize:
 * @object: The object to finalize
 **/
static void
egg_markdown_finalize (GObject *object)
{
	EggMarkdown *self;
	int i;

	g_return_if_fail (EGG_IS_MARKDOWN (object));

	self = EGG_MARKDOWN (object);

	g_return_if_fail (self->priv != NULL);
	g_string_free (self->priv->pending, TRUE);
	g_string_free (self->priv->processed, TRUE);

	for (i = 0; i < self->priv->link_table->len; i++) {
		g_free(g_array_index(self->priv->link_table, gchar *, i));
	}
	g_array_free (self->priv->link_table, TRUE);
	for (i = 0; i < self->priv->code_table->len; i++) {
		g_free(g_array_index(self->priv->code_table, gchar *, i));
	}
	g_array_free (self->priv->code_table, TRUE);
	g_free (self->priv->input_filename);

	for (i = 0; i < self->priv->regex_table->len; i++) {
		GRegex *p = g_array_index (self->priv->regex_table, GRegex *, i);
		if (p)
			g_regex_unref (p);
	}
	g_array_free (self->priv->regex_table, TRUE);

	G_OBJECT_CLASS (egg_markdown_parent_class)->finalize (object);
}

/**
 * egg_markdown_new:
 *
 * Return value: a new EggMarkdown object.
 **/
EggMarkdown *
egg_markdown_new (void)
{
	EggMarkdown *self;
	self = g_object_new (EGG_TYPE_MARKDOWN, NULL);
	return EGG_MARKDOWN (self);
}

/***************************************************************************
 ***                          MAKE CHECK TESTS                           ***
 ***************************************************************************/
#ifdef EGG_TEST
#include "egg-test.h"

void
egg_markdown_test (EggTest *test)
{
	EggMarkdown *self;
	gchar *text;
	gboolean ret;
	const gchar *markdown;
	const gchar *markdown_expected;

	if (!egg_test_start (test, "EggMarkdown"))
		return;

	/************************************************************
	 ****************        line_is_rule          **************
	 ************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("* * *");
	egg_test_title_assert (test, "is rule (1)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("***");
	egg_test_title_assert (test, "is rule (2)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("*****");
	egg_test_title_assert (test, "is rule (3)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("- - -");
	egg_test_title_assert (test, "is rule (4)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("---------------------------------------");
	egg_test_title_assert (test, "is rule (5)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("");
	egg_test_title_assert (test, "is rule (blank)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("richard hughes");
	egg_test_title_assert (test, "is rule (text)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_rule ("- richard-hughes");
	egg_test_title_assert (test, "is rule (bullet)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("");
	egg_test_title_assert (test, "is blank (blank)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank (" \t ");
	egg_test_title_assert (test, "is blank (mix)", ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("richard hughes");
	egg_test_title_assert (test, "is blank (name)", !ret);

	/************************************************************/
	ret = egg_markdown_to_text_line_is_blank ("ccccccccc");
	egg_test_title_assert (test, "is blank (full)", !ret);


	/************************************************************
	 ****************           replace            **************
	 ************************************************************/
	text = egg_markdown_replace ("summary -- really -- sure!", " -- ", " – ", NULL);
	egg_test_title (test, "replace (multiple)");
	if (g_str_equal (text, "summary – really – sure!"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************
	 ****************          formatter           **************
	 ************************************************************/
	text = egg_markdown_to_text_line_formatter ("**is important** text", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (left)");
	if (g_str_equal (text, "<b>is important</b> text"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this is **important**", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (right)");
	if (g_str_equal (text, "this is <b>important</b>"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("**important**", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (only)");
	if (g_str_equal (text, "<b>important</b>"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("***important***", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (only)");
	if (g_str_equal (text, "<b>*important</b>*"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("I guess * this is * not bold", "*", "<i>", "</i>");
	egg_test_title (test, "formatter (with spaces)");
	if (g_str_equal (text, "I guess * this is * not bold"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this **is important** text in **several** places", "**", "<b>", "</b>");
	egg_test_title (test, "formatter (middle, multiple)");
	if (g_str_equal (text, "this <b>is important</b> text in <b>several</b> places"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_word_auto_format_code_segment ("this is http://www.hughsie.com/with_spaces_in_url inline link");
	egg_test_title (test, "auto formatter (url)");
	if (g_str_equal (text, "this is `http://www.hughsie.com/with_spaces_in_url` inline link"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("this was \"triffic\" it was", "\"", "“", "”");
	egg_test_title (test, "formatter (quotes)");
	if (g_str_equal (text, "this was “triffic” it was"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************/
	text = egg_markdown_to_text_line_formatter ("This isn't a present", "'", "‘", "’");
	egg_test_title (test, "formatter (one quote)");
	if (g_str_equal (text, "This isn't a present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got %s", text);
	g_free (text);

	/************************************************************
	 ****************          markdown            **************
	 ************************************************************/
	egg_test_title (test, "get EggMarkdown object");
	self = egg_markdown_new ();
	egg_test_assert (test, self != NULL);

	/************************************************************/
	ret = egg_markdown_set_output (self, EGG_MARKDOWN_OUTPUT_PANGO);
	egg_test_title_assert (test, "set pango output", ret);

	/************************************************************/
	markdown = "OEMs\n"
		   "====\n"
		   " - Bullett\n";
	markdown_expected =
		   "<big>OEMs</big>\n"
		   "• Bullett";
	egg_test_title (test, "markdown (type2 header)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);


	/************************************************************/
	markdown = "this is http://www.hughsie.com/with_spaces_in_url inline link\n";
	markdown_expected = "this is <tt>http://www.hughsie.com/with_spaces_in_url</tt> inline link";
	egg_test_title (test, "markdown (autocode)");
	egg_markdown_set_autocode (self, TRUE);
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "*** This software is currently in alpha state ***\n";
	markdown_expected = "<b><i> This software is currently in alpha state </b></i>";
	egg_test_title (test, "markdown some invalid header");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = " - This is a *very*\n"
		   "   short paragraph\n"
		   "   that is not usual.\n"
		   " - Another";
	markdown_expected =
		   "• This is a <i>very</i> short paragraph that is not usual.\n"
		   "• Another";
	egg_test_title (test, "markdown (complex1)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "*  This is a *very*\n"
		   "   short paragraph\n"
		   "   that is not usual.\n"
		   "*  This is the second\n"
		   "   bullett point.\n"
		   "*  And the third.\n"
		   " \n"
		   "* * *\n"
		   " \n"
		   "Paragraph one\n"
		   "isn't __very__ long at all.\n"
		   "\n"
		   "Paragraph two\n"
		   "isn't much better.";
	markdown_expected =
		   "• This is a <i>very</i> short paragraph that is not usual.\n"
		   "• This is the second bullett point.\n"
		   "• And the third.\n"
		   "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n"
		   "Paragraph one isn't <b>very</b> long at all.\n"
		   "Paragraph two isn't much better.";
	egg_test_title (test, "markdown (complex1)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "This is a spec file description or\n"
		   "an **update** description in bohdi.\n"
		   "\n"
		   "* * *\n"
		   "# Big title #\n"
		   "\n"
		   "The *following* things 'were' fixed:\n"
		   "- Fix `dave`\n"
		   "* Fubar update because of \"security\"\n";
	markdown_expected =
		   "This is a spec file description or an <b>update</b> description in bohdi.\n"
		   "⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯⎯\n"
		   "<big>Big title</big>\n"
		   "The <i>following</i> things 'were' fixed:\n"
		   "• Fix <tt>dave</tt>\n"
		   "• Fubar update because of \"security\"";
	egg_test_title (test, "markdown (complex2)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	markdown = "* list seporated with spaces -\n"
		   "  first item\n"
		   "\n"
		   "* second item\n"
		   "\n"
		   "* third item\n";
	markdown_expected =
		   "• list seporated with spaces - first item\n"
		   "• second item\n"
		   "• third item";
	egg_test_title (test, "markdown (list with spaces)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	ret = egg_markdown_set_max_lines (self, 1);
	egg_test_title_assert (test, "set max_lines", ret);

	/************************************************************/
	markdown = "* list seporated with spaces -\n"
		   "  first item\n"
		   "* second item\n";
	markdown_expected =
		   "• list seporated with spaces - first item";
	egg_test_title (test, "markdown (one line limit)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	ret = egg_markdown_set_max_lines (self, 1);
	egg_test_title_assert (test, "set max_lines", ret);

	/************************************************************/
	markdown = "* list & spaces";
	markdown_expected =
		   "• list & spaces";
	egg_test_title (test, "markdown (escaping)");
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	/************************************************************/
	egg_test_title (test, "markdown (free text)");
	text = egg_markdown_parse (self, "This isn't a present");
	if (g_str_equal (text, "This isn't a present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s'", text);
	g_free (text);

	/************************************************************/
	egg_test_title (test, "markdown (autotext underscore)");
	text = egg_markdown_parse (self, "This isn't CONFIG_UEVENT_HELPER_PATH present");
	if (g_str_equal (text, "This isn't <tt>CONFIG_UEVENT_HELPER_PATH</tt> present"))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s'", text);
	g_free (text);

	/************************************************************/
	markdown = "*Thu Mar 12 12:00:00 2009* Dan Walsh <dwalsh@redhat.com> - 2.0.79-1\n"
		   "- Update to upstream \n"
		   " * Netlink socket handoff patch from Adam Jackson.\n"
		   " * AVC caching of compute_create results by Eric Paris.\n"
		   "\n"
		   "*Tue Mar 10 12:00:00 2009* Dan Walsh <dwalsh@redhat.com> - 2.0.78-5\n"
		   "- Add patch from ajax to accellerate X SELinux \n"
		   "- Update eparis patch\n";
	markdown_expected =
		   "<i>Thu Mar 12 12:00:00 2009</i> Dan Walsh <tt>&lt;dwalsh@redhat.com&gt;</tt> - 2.0.79-1\n"
		   "• Update to upstream\n"
		   "• Netlink socket handoff patch from Adam Jackson.\n"
		   "• AVC caching of compute_create results by Eric Paris.\n"
		   "<i>Tue Mar 10 12:00:00 2009</i> Dan Walsh <tt>&lt;dwalsh@redhat.com&gt;</tt> - 2.0.78-5\n"
		   "• Add patch from ajax to accellerate X SELinux\n"
		   "• Update eparis patch";
	egg_test_title (test, "markdown (end of bullett)");
	egg_markdown_set_escape (self, TRUE);
	ret = egg_markdown_set_max_lines (self, 1024);
	text = egg_markdown_parse (self, markdown);
	if (g_str_equal (text, markdown_expected))
		egg_test_success (test, NULL);
	else
		egg_test_failed (test, "failed, got '%s', expected '%s'", text, markdown_expected);
	g_free (text);

	g_object_unref (self);

	egg_test_end (test);
}
#endif

